from __future__ import annotations
import threading
import datetime
import uuid
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Union, Optional, TypeVar, List, Dict, TYPE_CHECKING

from ravendb.changes.database_changes import DatabaseChanges
from ravendb.documents.bulk_insert_operation import BulkInsertOperation, BulkInsertOptions
from ravendb.documents.indexes.index_creation import IndexCreation
from ravendb.documents.operations.executor import MaintenanceOperationExecutor, OperationExecutor
from ravendb.documents.operations.indexes import PutIndexesOperation
from ravendb.documents.session.event_args import (
    BeforeStoreEventArgs,
    AfterSaveChangesEventArgs,
    BeforeDeleteEventArgs,
    BeforeQueryEventArgs,
    SessionCreatedEventArgs,
    SessionClosingEventArgs,
    BeforeConversionToDocumentEventArgs,
    AfterConversionToDocumentEventArgs,
    BeforeConversionToEntityEventArgs,
    AfterConversionToEntityEventArgs,
    BeforeRequestEventArgs,
    SucceedRequestEventArgs,
    FailedRequestEventArgs,
)
from ravendb.documents.store.lazy import Lazy
from ravendb.documents.session.document_session import DocumentSession
from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
    InMemoryDocumentSessionOperations,
)
from ravendb.documents.session.misc import SessionOptions
from ravendb.documents.subscriptions.document_subscriptions import DocumentSubscriptions
from ravendb.documents.time_series import TimeSeriesOperations
from ravendb.http.request_executor import RequestExecutor
from ravendb.documents.identity.hilo import MultiDatabaseHiLoGenerator
from ravendb.http.topology import Topology
from ravendb.tools.utils import CaseInsensitiveDict
from ravendb.documents.conventions import DocumentConventions

T = TypeVar("T")

if TYPE_CHECKING:
    from ravendb.documents.indexes.abstract_index_creation_tasks import AbstractIndexCreationTask


class DocumentStoreBase:
    def __init__(self):
        self.__conventions = None
        self._initialized: bool = False

        self.__certificate_pem_path: Optional[str] = None
        self.__trust_store_path: Optional[str] = None

        self._urls: List[str] = []
        self._database: Optional[str] = None
        self._disposed: Optional[bool] = None

        self.__before_store: List[Callable[[BeforeStoreEventArgs], None]] = []
        self.__after_save_changes: List[Callable[[AfterSaveChangesEventArgs], None]] = []
        self.__before_delete: List[Callable[[BeforeDeleteEventArgs], None]] = []
        self.__before_query: List[Callable[[BeforeQueryEventArgs], None]] = []
        self.__on_session_creation: List[Callable[[SessionCreatedEventArgs], None]] = []
        self.__on_session_closing: List[Callable[[SessionClosingEventArgs], None]] = []

        self.__before_conversion_to_document: List[Callable[[BeforeConversionToDocumentEventArgs], None]] = []
        self.__after_conversion_to_document: List[Callable[[AfterConversionToDocumentEventArgs], None]] = []
        self.__before_conversion_to_entity: List[Callable[[BeforeConversionToEntityEventArgs], None]] = []
        self.__after_conversion_to_entity: List[Callable[[AfterConversionToEntityEventArgs], None]] = []

        self.__on_before_request: List[Callable[[BeforeRequestEventArgs], None]] = []
        self.__on_succeed_request: List[Callable[[SucceedRequestEventArgs], None]] = []

        self.__on_failed_request: List[Callable[[FailedRequestEventArgs], None]] = []
        self.__on_topology_updated: List[Callable[[Topology], None]] = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def disposed(self) -> bool:
        return self._disposed

    @property
    def conventions(self):
        if self.__conventions is None:
            self.__conventions = DocumentConventions()
        return self.__conventions

    @conventions.setter
    def conventions(self, value: DocumentConventions):
        self.__assert_not_initialized("conventions")
        self.__conventions = value

    @property
    def urls(self):
        return self._urls

    @urls.setter
    def urls(self, value: List[str]):
        self.__assert_not_initialized("urls")

        if value is None:
            raise ValueError("Value is None")

        for i in range(len(value)):
            if value[i] is None:
                raise ValueError("Urls cannot contain None")
            value[i] = str.rstrip(value[i], "/")

        self._urls = value

    @property
    def database(self) -> str:
        return self._database

    @database.setter
    def database(self, value: str):
        self.__assert_not_initialized("database")
        self._database = value

    @property
    def certificate_pem_path(self) -> str:
        return self.__certificate_pem_path

    @certificate_pem_path.setter
    def certificate_pem_path(self, value: str):
        self.__assert_not_initialized("certificate_url")
        if value.endswith("pfx"):
            raise ValueError("Invalid certificate format. " "Please use .pem file.")
        with open(value) as file:
            content = file.read()
            if "BEGIN CERTIFICATE" not in content:
                raise ValueError(
                    f"Invalid file. File stored under the path '{value}' isn't valid .pem certificate. "
                    f"BEGIN CERTIFICATE header wasn't found."
                )
            if "PRIVATE KEY" not in content:
                raise ValueError(
                    f"Invalid file. File stored under the path '{value}' isn't valid .pem certificate. "
                    f"PRIVATE KEY header wasn't found."
                )
        self.__certificate_pem_path = value

    @property
    def trust_store_path(self) -> str:
        return self.__trust_store_path

    @trust_store_path.setter
    def trust_store_path(self, value: str):
        self.__trust_store_path = value

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def maintenance(self) -> MaintenanceOperationExecutor:
        pass

    @abstractmethod
    def operations(self) -> OperationExecutor:
        pass

    # todo: aggressive_caching

    # todo: time_series

    @abstractmethod
    def open_session(self, database: Optional[str] = None, session_options: Optional = None):
        pass

    def add_on_session_creation(self, event: Callable[[SessionCreatedEventArgs], None]):
        self.__on_session_creation.append(event)

    def remove_on_session_creation(self, event: Callable[[SessionCreatedEventArgs], None]):
        self.__on_session_creation.remove(event)

    def add_on_session_closing(self, event: Callable[[SessionClosingEventArgs], None]):
        self.__on_session_closing.append(event)

    def remove_on_session_closing(self, event: Callable[[SessionClosingEventArgs], None]):
        self.__on_session_closing.remove(event)

    def add_before_conversion_to_document(self, event: Callable[[BeforeConversionToDocumentEventArgs], None]):
        self.__before_conversion_to_document.append(event)

    def remove_before_conversion_to_document(self, event: Callable[[BeforeConversionToDocumentEventArgs], None]):
        self.__before_conversion_to_document.remove(event)

    def add_after_conversion_to_document(self, event: Callable[[AfterConversionToDocumentEventArgs], None]):
        self.__after_conversion_to_document.append(event)

    def remove_after_conversion_to_document(self, event: Callable[[AfterConversionToDocumentEventArgs], None]):
        self.__after_conversion_to_document.remove(event)

    def add_before_conversion_to_entity(self, event: Callable[[BeforeConversionToEntityEventArgs], None]):
        self.__before_conversion_to_entity.append(event)

    def remove_before_conversion_to_entity(self, event: Callable[[BeforeConversionToEntityEventArgs], None]):
        self.__before_conversion_to_entity.remove(event)

    def add_after_conversion_to_entity(self, event: Callable[[AfterConversionToEntityEventArgs], None]):
        self.__after_conversion_to_entity.append(event)

    def remove_after_conversion_to_entity(self, event: Callable[[AfterConversionToEntityEventArgs], None]):
        self.__after_conversion_to_entity.remove(event)

    def add_on_before_request(self, event: Callable[[BeforeRequestEventArgs], None]):
        self.__on_before_request.append(event)

    def remove_on_before_request(self, event: Callable[[BeforeRequestEventArgs], None]):
        self.__on_before_request.remove(event)

    def add_on_succeed_request(self, event: Callable[[SucceedRequestEventArgs], None]):
        self.__on_succeed_request.append(event)

    def remove_on_succeed_request(self, event: Callable[[SucceedRequestEventArgs], None]):
        self.__on_succeed_request.remove(event)

    def add_on_failed_request(self, event: Callable[[FailedRequestEventArgs], None]):
        self.__on_failed_request.append(event)

    def remove_on_failed_request(self, event: Callable[[FailedRequestEventArgs], None]):
        self.__on_failed_request.remove(event)

    def register_events_for_session(self, session: InMemoryDocumentSessionOperations):
        for event in self.__before_store:
            session.add_before_store(event)

        for event in self.__after_save_changes:
            session.add_after_save_changes(event)

        for event in self.__before_delete:
            session.add_before_delete(event)

        for event in self.__before_query:
            session.add_before_query(event)

        for event in self.__before_conversion_to_document:
            session.add_before_conversion_to_document(event)

        for event in self.__after_conversion_to_document:
            session.add_after_conversion_to_document(event)

        for event in self.__before_conversion_to_entity:
            session.add_before_conversion_to_entity(event)

        for event in self.__after_conversion_to_entity:
            session.add_after_conversion_to_entity(event)

        for event in self.__on_session_closing:
            session.add_session_closing(event)

    def register_events_for_request_executor(self, request_executor: RequestExecutor):
        for event in self.__on_before_request:
            request_executor.add_on_before_request(event)
        for event in self.__on_failed_request:
            request_executor.add_on_failed_request(event)
        for event in self.__on_succeed_request:
            request_executor.add_on_succeed_request(event)
        for event in self.__on_topology_updated:
            request_executor.add_on_topology_updated(event)

    def after_session_created(self, session: InMemoryDocumentSessionOperations):
        for event in self.__on_session_creation:
            event(SessionCreatedEventArgs(session))

    def _ensure_not_closed(self) -> None:
        if self.disposed:
            raise ValueError("The document store has already been disposed and cannot be used")

    def __assert_not_initialized(self, property: str) -> None:
        if self._initialized:
            raise RuntimeError(f"You cannot set '{property}' after the document store has been initialized.")

    def assert_initialized(self) -> None:
        if not self._initialized:
            raise RuntimeError(
                f"You cannot open a session or access the database commands before initializing the "
                f"document store. Did you forget calling initialize()?"
            )

    def get_effective_database(self, database: str) -> str:
        return self.get_effective_database_static(self, database)

    @staticmethod
    def get_effective_database_static(store: DocumentStoreBase, database: str) -> str:
        if database is None:
            database = store.database

        if database and not database.isspace():
            return database

        raise ValueError(
            "Cannot determine database to operate on. "
            "Please either specify 'database' directly as an action parameter "
            "or set the default database to operate on using 'DocumentStore.database'. "
            "Did you forget to pass 'database' parameter?"
        )


class DocumentStore(DocumentStoreBase):
    def __init__(self, urls: Union[str, List[str]] = None, database: Optional[str] = None):
        super(DocumentStore, self).__init__()
        self.__subscriptions = DocumentSubscriptions(self)
        self.__thread_pool_executor = ThreadPoolExecutor()
        self.urls = [urls] if isinstance(urls, str) else urls
        self.database = database
        self.__request_executors: Dict[str, Lazy[RequestExecutor]] = CaseInsensitiveDict()
        # todo: aggressive cache
        self.__maintenance_operation_executor: Optional[MaintenanceOperationExecutor] = None
        self.__operation_executor: Optional[OperationExecutor] = None
        # todo: database smuggler
        self.__multi_db_hilo: Optional[MultiDatabaseHiLoGenerator] = None
        self.__identifier: Optional[str] = None
        self.__add_change_lock = threading.Lock()
        self.__database_changes: Dict[str, DatabaseChanges] = {}
        self.__after_close: List[Callable[[], None]] = []
        self.__before_close: List[Callable[[], None]] = []
        self.__time_series_operation: Optional[TimeSeriesOperations] = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def hilo_id_generator(self) -> Optional[MultiDatabaseHiLoGenerator]:
        return self.__multi_db_hilo

    @property
    def thread_pool_executor(self) -> ThreadPoolExecutor:
        return self.__thread_pool_executor

    @property
    def subscriptions(self) -> DocumentSubscriptions:
        return self.__subscriptions

    @property
    def identifier(self) -> Optional[str]:
        if self.__identifier is not None:
            return self.__identifier

        if self.urls is None:
            return None

        if self.database is not None:
            return ",".join(self.urls) + f" (DB: {self.database})"

        return ",".join(self.urls)

    @identifier.setter
    def identifier(self, value: str):
        self.__identifier = value

    def add_after_close(self, event: Callable[[], None]):
        self.__after_close.append(event)

    def remove_after_close(self, event: Callable[[], None]):
        self.__after_close.remove(event)

    def add_before_close(self, event: Callable[[], None]):
        self.__before_close.append(event)

    def remove_before_close(self, event: Callable[[], None]):
        self.__before_close.remove(event)

    def close(self):
        if self._disposed:
            return
        for event in self.__before_close:
            event()

        # todo: evict items from cache based on changes

        while len(self.__database_changes) > 0:
            self.__database_changes.popitem()[1].close()

        if self.__multi_db_hilo is not None:
            try:
                self.__multi_db_hilo.return_unused_range()
            except Exception:
                pass  # ignore

        if self.subscriptions is not None:
            self.subscriptions.close()

        for event in self.__after_close:
            event()

        for key, lazy in self.__request_executors.items():
            if not lazy.is_value_created:
                continue

            lazy.value.close()

        self.__thread_pool_executor.shutdown()
        self._disposed = True

    def open_session(
        self, database: Optional[str] = None, session_options: Optional[SessionOptions] = None
    ) -> DocumentSession:
        if not session_options:
            session_options = SessionOptions()
            session_options.database = database
            session_options.disable_atomic_document_writes_in_cluster_wide_transaction = (
                self.conventions.disable_atomic_document_writes_in_cluster_wide_transaction
            )

        self.assert_initialized()
        self._ensure_not_closed()

        session_id = uuid.uuid4()
        session = DocumentSession(self, session_id, session_options)
        self.register_events_for_session(session)
        self.after_session_created(session)
        return session

    def get_request_executor(self, database: Optional[str] = None) -> RequestExecutor:
        self.assert_initialized()

        database = self.get_effective_database(database)

        executor = self.__request_executors.get(database, None)
        if executor:
            return executor.value
        effective_database = database

        def __create_request_executor() -> RequestExecutor:
            request_executor = RequestExecutor.create(
                self.urls,
                effective_database,
                self.conventions,
                self.certificate_pem_path,
                self.trust_store_path,
                self.thread_pool_executor,
            )
            # todo: register events
            return request_executor

        def __create_request_executor_for_single_node() -> RequestExecutor:
            for_single_node = RequestExecutor.create_for_single_node_with_configuration_updates(
                self.urls[0],
                effective_database,
                self.conventions,
                self.certificate_pem_path,
                self.trust_store_path,
                self.thread_pool_executor,
            )
            # todo: register events

            return for_single_node

        executor = (
            Lazy(__create_request_executor)
            if not self.conventions.disable_topology_updates
            else Lazy(__create_request_executor_for_single_node)
        )

        self.__request_executors[database] = executor

        return executor.value

    def execute_index(self, task: "AbstractIndexCreationTask", database: Optional[str] = None) -> None:
        self.assert_initialized()
        task.execute(self, self.conventions, database)

    def execute_indexes(self, tasks: "List[AbstractIndexCreationTask]", database: Optional[str] = None) -> None:
        self.assert_initialized()
        indexes_to_add = IndexCreation.create_indexes_to_add(tasks, self.conventions)

        self.maintenance.for_database(self.get_effective_database(database)).send(PutIndexesOperation(*indexes_to_add))

    def changes(self, database=None, on_error=None, executor=None) -> DatabaseChanges:  # todo: sync with java
        self.assert_initialized()
        if not database:
            database = self.database
        with self.__add_change_lock:
            if database not in self.__database_changes:
                self.__database_changes[database] = DatabaseChanges(
                    request_executor=self.get_request_executor(database),
                    database_name=database,
                    on_close=self.__on_close_change,
                    on_error=on_error,
                    executor=executor,
                )
            return self.__database_changes[database]

    def __on_close_change(self, database_name: str):
        self.__database_changes.pop(database_name, None)

    def set_request_timeout(self, timeout: datetime.timedelta, database: Optional[str] = None) -> Callable[[], None]:
        self.assert_initialized()

        database = self.get_effective_database(database)

        request_executor = self.get_request_executor(database)
        old_timeout = request_executor.default_timeout
        request_executor.default_timeout = timeout

        def __return_function():
            request_executor.default_timeout = old_timeout

        return __return_function

    def initialize(self) -> DocumentStore:
        if self._initialized:
            return self

        self._assert_valid_configuration()

        RequestExecutor.validate_urls(self.urls)  # todo: pass also certificate

        if self.conventions.document_id_generator is None:  # don't overwrite what the user is doing
            generator = MultiDatabaseHiLoGenerator(self)
            self.__multi_db_hilo = generator

            self.conventions.document_id_generator = generator.generate_document_id

        self.conventions.freeze()
        self._initialized = True
        return self

    # todo: aggressively cache

    def bulk_insert(self, database_name: str = None, options: BulkInsertOptions = None) -> BulkInsertOperation:
        self.assert_initialized()
        return BulkInsertOperation(self.get_effective_database(database_name), self, options)

    def _assert_valid_configuration(self) -> None:
        if not self.urls:
            raise ValueError("Document URLs cannot be empty.")

    @property
    def maintenance(self) -> MaintenanceOperationExecutor:
        self.assert_initialized()

        if self.__maintenance_operation_executor is None:
            self.__maintenance_operation_executor = MaintenanceOperationExecutor(self)

        return self.__maintenance_operation_executor

    @property
    def operations(self) -> OperationExecutor:
        if self.__operation_executor is None:
            self.__operation_executor = OperationExecutor(self)

        return self.__operation_executor

    @property
    def time_series(self) -> TimeSeriesOperations:
        if self.__time_series_operation is None:
            self.__time_series_operation = TimeSeriesOperations(self)

        return self.__time_series_operation



from __future__ import annotations

from typing import Optional

from ravendb.primitives import constants


class MetadataAsDictionary:
    RESERVED_METADATA_PROPERTIES = {
        constants.Documents.Metadata.COLLECTION,
        constants.Documents.Metadata.ID,
        constants.Documents.Metadata.CHANGE_VECTOR,
        constants.Documents.Metadata.LAST_MODIFIED,
        constants.Documents.Metadata.RAVEN_PYTHON_TYPE,
    }

    def __init__(self, metadata=None, parent=None, parent_key=None):
        self._source = None if isinstance(metadata, dict) else metadata
        self._parent = parent
        self._parent_key = parent_key
        self._has_changes: Optional[bool] = None
        if metadata is None or len(metadata) == 0:
            self._metadata = {}
        else:
            self._initialize(metadata)

    def __len__(self):
        return len(self._metadata)

    def __setitem__(self, key, value):
        current_value = self._metadata.get(key, None)
        if current_value is None or current_value != value:
            self._metadata[key] = value
            self._has_changes = True
        return None

    def __delitem__(self, key):
        self._has_changes = True
        del self._metadata[key]

    def __contains__(self, item):
        return item in self._metadata

    def __getitem__(self, item):
        return self._metadata[item]

    def __getattr__(self, item):
        if item in ["update", "popitem"]:
            self.is_dirty = True
            return self._metadata.__getattribute__(item)
        elif item in ["pop", "clear"]:
            return super().__getattribute__(item)
        return self._metadata.__getattribute__(item)

    def clear(self) -> None:
        if self._metadata is None:
            self._initialize(self._source)

        keys_to_remove = set()

        for key, value in self._metadata.items():
            if key in self.RESERVED_METADATA_PROPERTIES:
                continue
            keys_to_remove.add(key)

        if keys_to_remove:
            for s in keys_to_remove:
                del self._metadata[s]
            self._has_changes = True

    def pop(self, key, default=None):  # todo: it'll never throw any exception, because default is always sent
        if self._metadata is None:
            self._initialize(self._source)
        if key in self._metadata:
            self._has_changes = True
        return self._metadata.pop(key, default)

    @property
    def is_dirty(self):
        return self._metadata is not None and self._has_changes is True

    @property
    def metadata(self):
        return self._metadata

    @is_dirty.setter
    def is_dirty(self, value):
        self._has_changes = value

    @property
    def has_parent(self):
        return self._parent is not None

    @classmethod
    def materialize_from_json(cls, metadata: dict) -> MetadataAsDictionary:
        result = cls(None)
        result._initialize(metadata)
        return result

    def _initialize(self, metadata_obj):
        # assign dictionary
        if isinstance(metadata_obj, dict):
            self._metadata = {}
            for key, value in metadata_obj.items():
                self._metadata[key] = self._convert_value(key, value)
            return

        # fetch object fields and put them into dict
        self._metadata = {}
        for key, value in metadata_obj.__dict__.items():
            self._metadata[key] = self._convert_value(key, value)

        if self._parent is not None:  # mark parent as dirty
            self._parent.update(self._parent_key, self)

        return self._metadata

    def _convert_value(self, key: str, value):
        if isinstance(value, (int, str, bool, float, type(None))):
            return value
        if hasattr(value, "__iter__") and type(value) is not dict:  # some collection
            return list(map(lambda x: self._convert_value(key, x), value))
        # not iterable, not simple type -> object or dictionary
        return MetadataAsDictionary(value, parent=self, parent_key=key)



from __future__ import annotations
from typing import Tuple, Union, List, Optional, Callable

import ravendb.documents.queries.facets.definitions as facets_definitions
from ravendb.documents.queries.facets.misc import FacetAggregation, FacetOptions
from ravendb.documents.session.tokens.query_tokens.query_token import QueryToken
from ravendb.tools.utils import Utils, QueryFieldUtil


class FacetToken(QueryToken):
    def __init__(
        self,
        aggregate_by_field_name: Optional[str] = None,
        alias: Optional[str] = None,
        ranges: Optional[List[str]] = None,
        options_parameter_name: Optional[str] = None,
    ):
        self._aggregate_by_field_name = aggregate_by_field_name
        self._alias = alias
        self._ranges = ranges
        self._options_parameter_name = options_parameter_name
        self._aggregations = []
        self._facet_setup_document_id: Optional[str] = None

    @classmethod
    def from_facet_setup_document_id(cls, facet_setup_document_id: str) -> FacetToken:
        token = cls()
        token._facet_setup_document_id = facet_setup_document_id
        return token

    @property
    def name(self) -> str:
        return self._alias or self._aggregate_by_field_name

    @classmethod
    def create(cls) -> FacetToken:
        pass

    @classmethod
    def create_from_facet_setup_document_id(
        cls,
        facet_setup_document_id: str = None,
    ) -> FacetToken:
        if facet_setup_document_id is None or facet_setup_document_id.isspace():
            raise ValueError("facet_setup_document_id cannot be None or whitespace")
        return FacetToken.from_facet_setup_document_id(facet_setup_document_id)

    @classmethod
    def create_from_facet(
        cls, facet: facets_definitions.Facet, add_query_parameter: Callable[[object], str]
    ) -> FacetToken:
        options_parameter_name = cls.__get_options_parameter_name(facet, add_query_parameter)
        token = FacetToken(
            QueryFieldUtil.escape_if_necessary(facet.field_name),
            QueryFieldUtil.escape_if_necessary(facet.display_field_name),
            None,
            options_parameter_name,
        )

        cls.__apply_aggregations(facet, token)

        return token

    @classmethod
    def create_from_range_facet(
        cls, facet: facets_definitions.RangeFacet, add_query_parameter: Callable[[object], str]
    ) -> FacetToken:
        options_parameter_name = cls.__get_options_parameter_name(facet, add_query_parameter)

        token = FacetToken(
            None, QueryFieldUtil.escape_if_necessary(facet.display_field_name), facet.ranges, options_parameter_name
        )

        cls.__apply_aggregations(facet, token)

        return token

    @classmethod
    def create_from_generic_range_facet(
        cls, facet: facets_definitions.GenericRangeFacet, add_query_parameter: Callable[[object], str]
    ) -> FacetToken:
        options_parameter_name = cls.__get_options_parameter_name(facet, add_query_parameter)

        ranges = [
            facets_definitions.GenericRangeFacet.parse(range_builder, add_query_parameter)
            for range_builder in facet.ranges
        ]

        token = FacetToken(
            None, QueryFieldUtil.escape_if_necessary(facet.display_field_name), ranges, options_parameter_name
        )

        cls.__apply_aggregations(facet, token)
        return token

    @classmethod
    def create_from_facet_base(
        cls, facet: facets_definitions.FacetBase, add_query_parameter: Callable[[object], str]
    ) -> FacetToken:
        # this is just a dispatcher
        return facet.to_facet_token(add_query_parameter)

    def write_to(self, writer: List[str]) -> None:
        writer.append("facet(")

        if self._facet_setup_document_id is not None:
            writer.append("id('")
            writer.append(self._facet_setup_document_id)
            writer.append("'))")
            return

        first_argument = False

        if self._aggregate_by_field_name is not None:
            writer.append(self._aggregate_by_field_name)
        elif self._ranges is not None:
            first_in_range = True

            for range_ in self._ranges:
                if not first_in_range:
                    writer.append(", ")

                first_in_range = False
                writer.append(range_)

        else:
            first_argument = True

        for aggregation in self._aggregations:
            if not first_argument:
                writer.append(", ")
            first_argument = False
            aggregation.write_to(writer)

        if self._options_parameter_name and not self._options_parameter_name.isspace():
            writer.append(", $")
            writer.append(self._options_parameter_name)

        writer.append(")")

        if not self._alias or self._alias.isspace() or self._alias == self._aggregate_by_field_name:
            return

        writer.append(" as ")
        writer.append(self._alias)

    @staticmethod
    def __get_options_parameter_name(
        facet: facets_definitions.FacetBase, add_query_parameter: Callable[[object], str]
    ) -> Optional[str]:
        return (
            add_query_parameter(facet.options)
            if facet.options is not None and facet.options != FacetOptions.default_options()
            else None
        )

    @classmethod
    def __apply_aggregations(cls, facet: facets_definitions.FacetBase, token: FacetToken) -> None:
        for key, value in facet.aggregations.items():
            for value_ in value:
                if key == FacetAggregation.MAX:
                    aggregation_token = cls._FacetAggregationToken.max(value_.name, value_.display_name)
                elif key == FacetAggregation.MIN:
                    aggregation_token = cls._FacetAggregationToken.min(value_.name, value_.display_name)
                elif key == FacetAggregation.AVERAGE:
                    aggregation_token = cls._FacetAggregationToken.average(value_.name, value_.display_name)
                elif key == FacetAggregation.SUM:
                    aggregation_token = cls._FacetAggregationToken.sum(value_.name, value_.display_name)
                else:
                    raise NotImplementedError(f"Unsupported aggregation method: {key}")

                token._aggregations.append(aggregation_token)

    class _FacetAggregationToken(QueryToken):
        def __init__(self, field_name: str, field_display_name: str, aggregation: FacetAggregation):
            self.__field_name = field_name
            self.__field_display_name = field_display_name
            self.__aggregation = aggregation

        def write_to(self, writer: List[str]) -> None:
            if self.__aggregation == FacetAggregation.MAX:
                writer.append("max(")
                writer.append(self.__field_name)
                writer.append(")")

            elif self.__aggregation == FacetAggregation.MIN:
                writer.append("min(")
                writer.append(self.__field_name)
                writer.append(")")

            elif self.__aggregation == FacetAggregation.AVERAGE:
                writer.append("avg(")
                writer.append(self.__field_name)
                writer.append(")")

            elif self.__aggregation == FacetAggregation.SUM:
                writer.append("sum(")
                writer.append(self.__field_name)
                writer.append(")")

            else:
                raise ValueError(f"Invalid aggregation mode: {self.__aggregation}")

            if not self.__field_display_name or self.__field_display_name.isspace():
                return

            writer.append(" as ")
            self.write_field(writer, self.__field_display_name)

        @classmethod
        def max(cls, field_name: str, field_display_name: Optional[str] = None) -> FacetToken._FacetAggregationToken:
            if field_name.isspace():
                raise ValueError("field_name cannot be None")
            return cls(field_name, field_display_name, FacetAggregation.MAX)

        @classmethod
        def min(cls, field_name: str, field_display_name: Optional[str] = None) -> FacetToken._FacetAggregationToken:
            if field_name.isspace():
                raise ValueError("field_name cannot be None")
            return cls(field_name, field_display_name, FacetAggregation.MIN)

        @classmethod
        def average(
            cls, field_name: str, field_display_name: Optional[str] = None
        ) -> FacetToken._FacetAggregationToken:
            if field_name.isspace():
                raise ValueError("field_name cannot be None")
            return cls(field_name, field_display_name, FacetAggregation.AVERAGE)

        @classmethod
        def sum(cls, field_name: str, field_display_name: Optional[str] = None) -> FacetToken._FacetAggregationToken:
            if field_name.isspace():
                raise ValueError("field_name cannot be None")
            return cls(field_name, field_display_name, FacetAggregation.SUM)



import datetime
import enum
import inspect
import logging
from typing import Union, Optional, TypeVar, List, Type, Callable, Dict, TYPE_CHECKING

from ravendb.primitives import constants
from ravendb.documents.commands.query import QueryCommand
from ravendb.documents.queries.index_query import IndexQuery
from ravendb.documents.queries.query import QueryResult
from ravendb.documents.session.event_args import BeforeConversionToEntityEventArgs, AfterConversionToEntityEventArgs
from ravendb.documents.session.tokens.query_tokens.definitions import FieldsToFetchToken
from ravendb.exceptions.documents.indexes import IndexDoesNotExistException
from ravendb.tools.utils import Stopwatch, Utils

if TYPE_CHECKING:
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )
    from ravendb.documents.session.document_session import DocumentSession


class QueryOperation:
    __logger = logging.getLogger("QueryOperation")
    _T = TypeVar("_T")

    # __facet_result_fields =
    def __init__(
        self,
        session: "DocumentSession",
        index_name: str,
        index_query: IndexQuery,
        fields_to_fetch: FieldsToFetchToken,
        disable_entities_tracking: bool,
        metadata_only: bool,
        index_entries_only: bool,
        is_project_into: bool,
    ):
        self.__session = session
        self.__index_name = index_name
        self.__index_query = index_query
        self.__fields_to_fetch = fields_to_fetch
        self.__no_tracking = disable_entities_tracking
        self.__metadata_only = metadata_only
        self.__index_entries_only = index_entries_only
        self.__is_project_into = is_project_into

        self.__current_query_results: Union[None, QueryResult] = None
        self.__sp: Union[None, Stopwatch] = Stopwatch()
        self.__assert_page_size_set()

    def create_request(self) -> QueryCommand:
        self.__session.increment_requests_count()
        self.log_query()
        return QueryCommand(self.__session, self.__index_query, self.__metadata_only, self.__index_entries_only)

    @property
    def current_query_results(self) -> QueryResult:
        return self.__current_query_results

    def set_result(self, query_result: QueryResult):
        self.ensure_is_acceptable_and_save_result(query_result)

    def __assert_page_size_set(self) -> None:
        if not self.__session.conventions.throw_if_query_page_size_is_not_set:
            return

        if self.__index_query.is_page_size_set:
            return

        raise RuntimeError(
            "Attempt to query without explicitly specifying a page size. "
            "You can use .take() methods to set maximum number of results. By default the page size is "
            "set to 0x7fffff and can cause severe performance degradation."
        )

    def __start_timing(self) -> None:
        self.__sp = Stopwatch().start()

    def log_query(self) -> None:
        self.__logger.debug(
            f"Executing query {self.__index_query.query} "
            f"on index {self.__index_name} "
            f"in {self.__session.advanced.store_identifier}"
        )

    def enter_query_context(self) -> None:  # todo: make it return Closeable
        self.__start_timing()

    # def complete_as_array(self, object_type: Type[_T]) -> List[_T]:
    #    query_result = self.__current_query_results.create_snapshot()
    #    result: List[_T] = [None]
    #    self.complete_internal(object_type, query_result, lambda idx, item: resource)

    def complete(self, object_type: Type[_T]) -> List[_T]:
        query_result = self.__current_query_results.create_snapshot()
        result = list()
        self.__complete_internal(object_type, query_result, result.insert)
        return result

    def __complete_internal(
        self,
        object_type: Type[_T],
        query_result: QueryResult[List[Dict[str, Dict[str, str]]], None],
        add_to_result: Callable[[int, _T], None],
    ) -> None:
        if not self.__no_tracking:
            self.__session.register_includes(query_result.includes)

        for i in range(len(query_result.results)):
            document = query_result.results[i]
            metadata = document.get(constants.Documents.Metadata.KEY)
            id_node = metadata.get(constants.Documents.Metadata.ID)
            key = None
            if isinstance(id_node, str):
                key = id_node

            add_to_result(
                i,
                self.deserialize(
                    object_type,
                    key,
                    document,
                    metadata,
                    self.__fields_to_fetch,
                    self.__no_tracking,
                    self.__session,
                    self.__is_project_into,
                ),
            )
        if not self.__no_tracking:
            self.__session.register_missing_includes(
                query_result.results, query_result.includes, query_result.included_paths
            )

            if query_result.counter_includes is not None:
                self.__session.register_counters(query_result.counter_includes, query_result.included_counter_names)

            if query_result.time_series_includes is not None:
                self.__session.register_time_series(query_result.time_series_includes)

            if query_result.compare_exchange_value_includes is not None:
                self.__session.cluster_transaction.register_compare_exchange_values(
                    query_result.compare_exchange_value_includes
                )

    @staticmethod
    def deserialize(
        object_type: Type[_T],
        key: str,
        document: dict,
        metadata: dict,
        fields_to_fetch: FieldsToFetchToken,
        disable_entities_tracking: bool,
        session: "InMemoryDocumentSessionOperations",
        is_project_into: bool,
    ):
        projection = metadata.get("@projection")
        if not projection:
            return session.track_entity(object_type, key, document, metadata, disable_entities_tracking)

        if (
            fields_to_fetch is not None
            and fields_to_fetch.projections is not None
            and len(fields_to_fetch.projections) == 1
        ):
            projection_field = fields_to_fetch.projections[0]

            if fields_to_fetch.source_alias is not None:
                if projection_field.startswith(fields_to_fetch.source_alias):
                    projection_field = projection_field[len(fields_to_fetch.source_alias) + 1 :]

                if projection_field.startswith("'"):
                    projection_field = projection_field[1:-1]

            if object_type == str or object_type in Utils.primitives_and_collections or object_type == enum.Enum:
                json_node = document.get(projection_field)
                if isinstance(json_node, object_type):
                    return json_node
                raise TypeError(f'The value under projection field "{projection_field}" isn\'t a "{object_type}"')

            # todo: time series
            # is_time_series_field = fields_to_fetch.projections[0].startswith(constants.TimeSeries.QUERY_FUNCTION)

            if not is_project_into:  # or is_time_series_field
                inner = document.get(projection_field)
                if inner is None:
                    return Utils.get_default_value(object_type)

                if (
                    fields_to_fetch.fields_to_fetch is not None
                    and fields_to_fetch.fields_to_fetch[0] == fields_to_fetch.projections[0]
                ):
                    document = inner

        if object_type == dict:
            return document

        session.before_conversion_to_entity_invoke(
            BeforeConversionToEntityEventArgs(session, key, object_type, document)
        )

        # By custom defined 'from_json' serializer class method
        # todo: make separate interface to do from_json
        if "from_json" in object_type.__dict__ and inspect.ismethod(object_type.from_json):
            result = object_type.from_json(document)
        else:
            result = Utils.initialize_object(document, object_type)
        session.after_conversion_to_entity_invoke(AfterConversionToEntityEventArgs(session, key, document, result))

        return result

    @property
    def no_tracking(self) -> bool:
        return self.__no_tracking

    @no_tracking.setter
    def no_tracking(self, value: bool):
        self.__no_tracking = value

    def ensure_is_acceptable_and_save_result(
        self, result: QueryResult, duration: Optional[datetime.timedelta] = None
    ) -> None:
        if duration is not None:
            if result is None:
                raise IndexDoesNotExistException(f"Could not find index {self.__index_name}")
            self.ensure_is_acceptable(result, self.__index_query.wait_for_non_stale_results, duration, self.__session)

            self.__save_query_result(result)

        elif self.__sp is None:
            self.ensure_is_acceptable_and_save_result(result, None)
        else:
            self.__sp.stop()
            self.ensure_is_acceptable_and_save_result(result, self.__sp.elapsed())

    def __save_query_result(self, result: QueryResult) -> None:
        self.__current_query_results = result

        is_stale = " stale " if result.is_stale else " "
        parameters = []
        if self.__index_query.query_parameters:
            parameters.append("(parameters: ")

            first = True

            for key, value in self.__index_query.query_parameters.items():
                if not first:
                    parameters.append(", ")

                parameters.append(key)
                parameters.append(" = ")
                parameters.append(value)

                first = False

            parameters.append(") ")

        self.__logger.debug(
            f"Query {self.__index_query.query} {''.join(list(map(str,parameters)))}returned "
            f"{len(result.results)}{is_stale}results (total index results: {result.total_results})"
        )

    @staticmethod
    def ensure_is_acceptable(
        result: QueryResult,
        wait_for_non_stale_results: bool,
        duration: Union[datetime.timedelta, Stopwatch],
        session: "InMemoryDocumentSessionOperations",
    ):
        if isinstance(duration, datetime.timedelta):
            if wait_for_non_stale_results and result.is_stale:
                elapsed = "" if duration is None else f" {(duration.microseconds / 1000)} ms"
                msg = f"Waited {elapsed} for the query to return non stale result."
                raise TimeoutError(msg)

    @property
    def index_query(self) -> IndexQuery:
        return self.__index_query



import os
from datetime import datetime
from typing import Optional

from ravendb import DocumentStore, GetIndexErrorsOperation


class RavenTestHelper:
    @staticmethod
    def utc_today() -> datetime:
        today = datetime.today()
        return datetime(today.year, today.month, today.day, 0, 0, 0, 0)

    @staticmethod
    def utc_this_month() -> datetime:
        today = datetime.today()
        return datetime(today.year, today.month, 1, 0, 0, 0, 0)

    @staticmethod
    def assert_no_index_errors(store: DocumentStore, database_name: Optional[str] = None) -> None:
        errors = store.maintenance.for_database(database_name).send(GetIndexErrorsOperation())

        sb = []
        for index_errors in errors:
            if not index_errors or not index_errors.errors:
                continue

            sb.append("Index Errors for '")
            sb.append(index_errors.name)
            sb.append("' (")
            sb.append(len(index_errors.errors))
            sb.append(")")
            sb.append(os.linesep)

            for index_error in index_errors.errors:
                sb.append(f"- {index_error}")
                sb.append(os.linesep)

            sb.append(os.linesep)

        if not sb:
            return

        raise RuntimeError("".join(map(str, sb)))



from __future__ import annotations

import abc
import datetime
import inspect
import math
from enum import Enum
from typing import List, Dict, Type, Tuple, Any, TypeVar, Generic, Optional

from ravendb.primitives import constants
from ravendb.exceptions.raven_exceptions import RavenException
from ravendb.primitives.time_series import TimeValue
from ravendb.tools.utils import Utils

_T_TSBindable = TypeVar("_T_TSBindable")
_T_Values = TypeVar("_T_Values")


class AbstractTimeSeriesRange(abc.ABC):
    def __init__(self, name: str):
        self.name = name


class TimeSeriesRange(AbstractTimeSeriesRange):
    def __init__(self, name: str, from_date: Optional[datetime.datetime], to_date: Optional[datetime.datetime]):
        super().__init__(name)
        self.from_date = from_date
        self.to_date = to_date


class TimeSeriesTimeRange(AbstractTimeSeriesRange):
    def __init__(self, name: str, time: TimeValue, type: TimeSeriesRangeType):
        super().__init__(name)
        self.time = time
        self.type = type


class TimeSeriesCountRange(AbstractTimeSeriesRange):
    def __init__(self, name: str, count: int, type: TimeSeriesRangeType):
        super().__init__(name)
        self.count = count
        self.type = type


class TimeSeriesRangeType(Enum):
    NONE = "None"
    LAST = "Last"


class TypedTimeSeriesRollupEntry(Generic[_T_Values]):
    def __init__(self, object_type: Type[_T_Values], timestamp: datetime.datetime):
        self._object_type = object_type
        self.tag: Optional[str] = None
        self.rollup = True
        self.timestamp = timestamp

        self._first: Optional[_T_Values] = None
        self._last: Optional[_T_Values] = None
        self._max: Optional[_T_Values] = None
        self._min: Optional[_T_Values] = None
        self._sum: Optional[_T_Values] = None
        self._count: Optional[_T_Values] = None
        self._average: Optional[_T_Values] = None

    def _create_instance(self) -> _T_Values:
        try:
            return Utils.try_get_new_instance(self._object_type)  # create object type instance
        except Exception as e:
            raise RavenException(f"Unable to create instance of class: {self._object_type.__name__}", e)

    @property
    def max(self) -> _T_Values:
        if self._max is None:
            self._max = self._create_instance()
        return self._max

    @property
    def min(self) -> _T_Values:
        if self._min is None:
            self._min = self._create_instance()
        return self._min

    @property
    def count(self) -> _T_Values:
        if self._count is None:
            self._count = self._create_instance()
        return self._count

    @property
    def first(self) -> _T_Values:
        if self._first is None:
            self._first = self._create_instance()
        return self._first

    @property
    def last(self) -> _T_Values:
        if self._last is None:
            self._last = self._create_instance()
        return self._last

    @property
    def sum(self) -> _T_Values:
        if self._sum is None:
            self._sum = self._create_instance()
        return self._sum

    @property
    def average(self) -> _T_Values:
        if self._average is not None:
            return self._average

        values_count = len(TimeSeriesValuesHelper.get_fields_mapping(self._object_type))

        sums = TimeSeriesValuesHelper.get_values(self._object_type, self._sum)
        counts = TimeSeriesValuesHelper.get_values(self._object_type, self._count)
        averages: List[Optional[float]] = [None] * values_count

        for i in range(values_count):
            if counts[i] < constants.min_normal:
                averages[i] = constants.nan_value
            else:
                averages[i] = sums[i] / counts[i]

        self._average = TimeSeriesValuesHelper.set_fields(self._object_type, averages)

        return self._average

    def get_values_from_members(self) -> List[float]:
        values_count = len(TimeSeriesValuesHelper.get_fields_mapping(self._object_type))

        result: List[Optional[float]] = [None] * values_count * 6
        self._assign_rollup(result, self._first, 0)
        self._assign_rollup(result, self._last, 1)
        self._assign_rollup(result, self._min, 2)
        self._assign_rollup(result, self._max, 3)
        self._assign_rollup(result, self._sum, 4)
        self._assign_rollup(result, self._count, 5)

        return result

    def _assign_rollup(self, target: List[float], source: _T_Values, offset: int) -> None:
        if source is not None:
            values = TimeSeriesValuesHelper.get_values(self._object_type, source)
            if values is not None:
                for i in range(len(values)):
                    target[i * 6 + offset] = values[i]

    @classmethod
    def from_entry(
        cls, ts_bindable_object_type: Type[_T_TSBindable], entry: TimeSeriesEntry
    ) -> TypedTimeSeriesRollupEntry[_T_TSBindable]:
        result = TypedTimeSeriesRollupEntry(ts_bindable_object_type, entry.timestamp)
        result.rollup = True
        result.tag = entry.tag

        values = entry.values

        result._first = TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, cls.extract_values(values, 0))
        result._last = TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, cls.extract_values(values, 1))
        result._min = TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, cls.extract_values(values, 2))
        result._max = TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, cls.extract_values(values, 3))
        result._sum = TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, cls.extract_values(values, 4))
        result._count = TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, cls.extract_values(values, 5))

        return result

    @staticmethod
    def extract_values(input: List[float], offset: int) -> List[float]:
        length = math.ceil((len(input) - offset) / 6.0)
        idx = 0
        result: List[Optional[float]] = [None] * length

        while idx < length:
            result[idx] = input[offset + idx * 6]
            idx += 1

        return result


class TypedTimeSeriesEntry(Generic[_T_TSBindable]):
    def __init__(
        self,
        timestamp: datetime.datetime = None,
        tag: str = None,
        values: List[int] = None,
        is_rollup: bool = None,
        value: _T_TSBindable = None,
    ):
        self.timestamp = timestamp
        self.tag = tag
        self.values = values
        self.is_rollup = is_rollup
        self.value = value

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]):
        return cls(
            Utils.string_to_datetime(json_dict["Timestamp"]),
            json_dict["Tag"],
            json_dict["Values"],
            json_dict["IsRollup"],
        )


class TimeSeriesEntry:
    def __init__(
        self, timestamp: datetime.datetime = None, tag: str = None, values: List[int] = None, rollup: bool = None
    ):
        self.timestamp = timestamp
        self.tag = tag
        self.values = values
        self.rollup = rollup

    @property
    def value(self):
        if len(self.values) == 1:
            return self.values[0]
        raise ValueError("Entry has more than one value.")

    @value.setter
    def value(self, value: int):
        if len(self.values) == 1:
            self.values[0] = value
            return
        raise ValueError("Entry has more than one value")

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]):
        return cls(
            Utils.string_to_datetime(json_dict["Timestamp"]),
            json_dict["Tag"],
            json_dict["Values"],
            json_dict["IsRollup"],
        )

    def as_typed_entry(self, ts_bindable_object_type: Type[_T_TSBindable]) -> TypedTimeSeriesEntry[_T_TSBindable]:
        return TypedTimeSeriesEntry(
            self.timestamp,
            self.tag,
            self.values,
            self.rollup,
            TimeSeriesValuesHelper.set_fields(ts_bindable_object_type, self.values, self.rollup),
        )


class ITimeSeriesValuesBindable(abc.ABC):
    @abc.abstractmethod
    def get_time_series_mapping(self) -> Dict[int, Tuple[str, Optional[str]]]:
        # return a dictionary {index of time series value - (name of 'float' field, label)}
        # e.g. return {0 : ('heart', 'Heartrate'), 1: ('bp', 'Blood Pressure')}
        # for some class that has 'heart' and 'bp' float fields
        raise NotImplementedError()


class TimeSeriesValuesHelper:
    _cache: Dict[Type, Dict[int, Tuple[str, str]]] = {}

    @staticmethod
    def create_ts_bindable_class_instance(ts_bindable_object_type: Type[_T_TSBindable]) -> _T_TSBindable:
        init_signature = inspect.signature(ts_bindable_object_type.__init__)
        init_parameters = init_signature.parameters.values()

        # Create a dictionary of arguments with default values
        arguments = {
            param.name: param.default if param.default != param.empty else None
            for param in init_parameters
            if param.name != "self"
        }

        instance = None

        try:
            instance = ts_bindable_object_type(**arguments)
        except Exception as e:
            raise TypeError(
                f"Failed to get time series fields mapping. "
                f"Failed to resolve the class instance fields, "
                f"initializing object of class '{ts_bindable_object_type.__name__}' using default args: '{arguments}'. "
                f"Is the {ts_bindable_object_type.__name__}__init__() raising errors in some cases?",
                e,
            )
        return instance

    @staticmethod
    def get_fields_mapping(ts_bindable_object_type: Type[_T_TSBindable]) -> Optional[Dict[int, Tuple[str, str]]]:
        if ts_bindable_object_type not in TimeSeriesValuesHelper._cache:
            if not issubclass(ts_bindable_object_type, ITimeSeriesValuesBindable):
                return None

            ts_bindable_class_instance = TimeSeriesValuesHelper.create_ts_bindable_class_instance(
                ts_bindable_object_type
            )
            ts_bindable_type_mapping = ts_bindable_class_instance.get_time_series_mapping()

            mappings = ts_bindable_type_mapping
            mapping_keys = list(ts_bindable_type_mapping.keys())
            if mapping_keys[0] != 0 or mapping_keys[-1] != len(mapping_keys) - 1:
                raise RuntimeError(
                    f"The mapping of '{ts_bindable_object_type.__name__ }' "
                    f"must contain consecutive values starting from 0."
                )

            for key, (name, tag) in mappings.items():
                class_instance_value = ts_bindable_class_instance.__dict__[name]
                if not isinstance(class_instance_value, float) and class_instance_value is not None:
                    raise RuntimeError(
                        f"Cannot create a mapping for '{ts_bindable_object_type.__name__}' class, "
                        f"because field '{name}' is not a float"
                    )

            new_cache_entry = {}
            for idx, field_name_and_ts_value_name in ts_bindable_type_mapping.items():
                field_name = field_name_and_ts_value_name[0]
                time_series_value_name = field_name_and_ts_value_name[1] or field_name
                new_cache_entry[idx] = (field_name, time_series_value_name)

            TimeSeriesValuesHelper._cache[ts_bindable_object_type] = new_cache_entry
        return TimeSeriesValuesHelper._cache[ts_bindable_object_type]

    @staticmethod
    def get_values(ts_bindable_object_type: Type[_T_TSBindable], obj: _T_TSBindable) -> Optional[List[float]]:
        mapping = TimeSeriesValuesHelper.get_fields_mapping(ts_bindable_object_type)
        if mapping is None:
            return None

        try:
            values: List[Optional[float]] = [None] * len(mapping)
            for key, value in mapping.items():
                values[key] = obj.__dict__[value[0]]  # get field value

            return values
        except Exception as e:
            raise RavenException("Unable to read time series values", e)

    @staticmethod
    def set_fields(
        ts_bindable_object_type: Type[_T_TSBindable], values: List[float], as_rollup: bool = False
    ) -> _T_TSBindable:
        if values is None:
            return None

        mapping = TimeSeriesValuesHelper.get_fields_mapping(ts_bindable_object_type)
        if mapping is None:
            return None

        try:
            ts_bindable_class_instance = TimeSeriesValuesHelper.create_ts_bindable_class_instance(
                ts_bindable_object_type
            )
            for index, field_name_and_ts_name_tuple in mapping.items():
                value = None  # Not-a-Number

                if index < len(values):
                    if as_rollup:
                        index *= 6
                    value = values[index]

                field_name = field_name_and_ts_name_tuple[0]
                ts_bindable_class_instance.__dict__[field_name] = value

            return ts_bindable_class_instance

        except Exception as e:
            raise RavenException("Unable to read time series values.", e)



from __future__ import annotations

import datetime
import http
from http import HTTPStatus
from abc import abstractmethod
from typing import Union, Optional, Callable, Generic, TypeVar, Dict, Type
from enum import Enum

import requests
from ravendb.extensions.http_extensions import HttpExtensions
from ravendb.http.http_cache import HttpCache
from ravendb.http.misc import ResponseDisposeHandling
from ravendb.http.server_node import ServerNode


class RavenCommandResponseType(Enum):
    EMPTY = "Empty"
    OBJECT = "Object"
    RAW = "Raw"

    def __str__(self):
        return self.value


_T_Result = TypeVar("_T_Result")

# todo: check what's wrong with this generic. it doesnt work e.g. in HiloCommand


class RavenCommand(Generic[_T_Result]):
    @classmethod
    def from_copy(cls, copy: RavenCommand[_T_Result]) -> RavenCommand[_T_Result]:
        command = cls(copy._result_class)
        command._response_type = copy.response_type
        command._can_cache = copy.can_cache
        command._can_cache_aggressively = copy.can_cache_aggressively
        command._selected_node_tag = copy.selected_node_tag
        return command

    def __init__(self, result_class: Type[_T_Result] = None):
        self._result_class = result_class
        self._response_type = RavenCommandResponseType.OBJECT
        self._can_cache_aggressively = True
        self._can_cache = True
        self.failover_topology_etag = -2

        self.result: Optional[_T_Result] = None
        self.status_code: Optional[int] = None
        self.timeout: Optional[datetime.timedelta] = None
        self._selected_node_tag: Optional[str] = None
        self._number_of_attempts: Optional[int] = None
        self.failed_nodes: Dict[ServerNode, Exception] = {}
        self.on_response_failure: Callable[[requests.Response], None] = lambda resp: None

    @abstractmethod
    def is_read_request(self) -> bool:
        pass

    @abstractmethod
    def create_request(self, node: ServerNode) -> requests.Request:
        pass

    @property
    def response_type(self) -> RavenCommandResponseType:
        return self._response_type

    @property
    def can_cache(self) -> bool:
        return self._can_cache

    @property
    def can_cache_aggressively(self) -> bool:
        return self._can_cache_aggressively

    @property
    def selected_node_tag(self) -> Optional[str]:
        return self._selected_node_tag

    @property
    def number_of_attempts(self) -> int:
        if self._number_of_attempts is None:
            self._number_of_attempts = 0
        return self._number_of_attempts

    @number_of_attempts.setter
    def number_of_attempts(self, value: int):
        self._number_of_attempts = value

    @abstractmethod
    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if self._response_type == RavenCommandResponseType.EMPTY or RavenCommandResponseType.RAW:
            self._throw_invalid_response()
        raise RuntimeError(
            f"{self.response_type.name} command must override the set_response method which "
            f"expects response with the following type {self.response_type}"
        )

    def send(self, session: requests.Session, request: requests.Request) -> requests.Response:
        return session.request(
            request.method,
            url=request.url,
            data=request.data,
            files=request.files,
            cert=session.cert,
            headers=request.headers,
        )

    def set_response_raw(self, response: requests.Response, stream: bytes) -> None:
        raise RuntimeError(
            f"When {self.response_type} is set to Raw then please override this method to handle the response "
        )

    @staticmethod
    def ensure_is_not_null_or_string(value: str, name: str) -> None:
        if not value:
            raise ValueError(f"{name} cannot be None or empty")

    def is_failed_with_node(self, node: ServerNode) -> bool:
        return self.failed_nodes and node in self.failed_nodes

    def process_response(self, cache: HttpCache, response: requests.Response, url) -> ResponseDisposeHandling:
        # todo: check if response is a dict here from beginning
        if not response:
            return ResponseDisposeHandling.AUTOMATIC

        if self.response_type == RavenCommandResponseType.EMPTY or response.status_code == HTTPStatus.NO_CONTENT:
            return ResponseDisposeHandling.AUTOMATIC

        try:
            if self.response_type == RavenCommandResponseType.OBJECT:
                content_length = len(response.content)
                if content_length == 0:
                    response.close()
                    return ResponseDisposeHandling.AUTOMATIC

                json_content = response.content.decode("utf-8")
                if cache is not None:
                    self._cache_response(cache, url, response, json_content)
                self.set_response(json_content, False)
                return ResponseDisposeHandling.AUTOMATIC
            else:
                self.set_response_raw(response, response.content)
        except Exception as e:
            raise e
        finally:
            response.close()
        return ResponseDisposeHandling.AUTOMATIC

    def _cache_response(self, cache: HttpCache, url: str, response: requests.Response, response_json: str) -> None:
        if not self.can_cache:
            return

        change_vector = HttpExtensions.get_etag_header(response)
        if change_vector is None:
            return
        cache.set(url, change_vector, response_json)

    @staticmethod
    def _throw_invalid_response(cause: Optional[BaseException] = None) -> None:
        raise ValueError(f"Response is invalid{f': {cause.args[0]}' if cause else ''}")

    def _add_change_vector_if_not_none(self, change_vector: str, request: requests.Request):
        if change_vector:
            if request.headers is not None:
                request.headers["If-Match"] = f'"{change_vector}"'
            else:
                request.headers = {"If-Match": f'"{change_vector}'}


class VoidRavenCommand(RavenCommand[None]):
    def __init__(self):
        super().__init__(None)
        self._response_type = RavenCommandResponseType.EMPTY

    @abstractmethod
    def create_request(self, node: ServerNode) -> requests.Request:
        pass

    def is_read_request(self) -> bool:
        return False

    def set_response(self, response: str, from_cache: bool) -> None:
        pass



from __future__ import annotations
from concurrent.futures import Future, ThreadPoolExecutor
from threading import Lock
from typing import Callable, Generic, TypeVar, Optional

from ravendb.tools.concurrentset import ConcurrentSet

_T_Change = TypeVar("_T_Change")


class Observable(Generic[_T_Change]):
    def __init__(
        self,
        on_connect: Optional[Callable[[], None]] = None,
        on_disconnect: Optional[Callable[[], None]] = None,
        executor: Optional[ThreadPoolExecutor] = None,
    ):
        self.on_connect = on_connect
        self._on_disconnect = on_disconnect
        self.last_exception = None
        self.value_lock = Lock()
        self._value = 0
        self._future = None
        self._future_set = Future()
        self._subscribers = ConcurrentSet()
        self._executor = executor

    def __call__(self, filter_method):
        self._filter = filter_method
        return self

    def on_document_change_notification(self, value: _T_Change):
        try:
            if self._filter(value):
                for subscriber in self._subscribers:
                    subscriber.on_next(value)
        except Exception as e:
            self.error(e)

    def subscribe(self, on_next_callback: Callable[[], None]) -> Callable[[], None]:
        """
        @param func on_next_callback: The action that observer will do when changes happens
        :return: method that close the subscriber
        """
        self.inc()
        observer = ActionObserver(on_next=on_next_callback)
        self._subscribers.add(observer)

        def close_action() -> None:
            self.dec()
            self._subscribers.remove(observer)
            if observer.on_completed_callback is not None:
                observer.on_completed()

        return close_action

    def subscribe_with_observer(self, observer: ActionObserver) -> Callable[[], None]:
        """
        @param Observer observer: The observer that will do action when changes happens
        :return: method that close the subscriber
        """
        self.inc()
        self._subscribers.add(observer)

        def close_action() -> None:
            self.dec()
            self._subscribers.remove(observer)
            if observer.on_completed_callback is not None:
                observer.on_completed()

        return close_action

    def inc(self):
        with self.value_lock:
            self._value += 1

    def dec(self):
        with self.value_lock:
            self._value -= 1
            if self._value == 0:
                self.set(self._executor.submit(self._on_disconnect))

    def set(self, future: Future) -> None:
        if not self._future_set.done():

            def done_callback(f):
                try:
                    if f.cancelled():
                        self._future_set.cancel()
                    elif f.exception():
                        self._future_set.set_exception(f.exception)
                    else:
                        self._future_set.set_result(None)
                except Exception:
                    raise

            future.add_done_callback(done_callback)
        self._future = future

    def error(self, exception: Exception):
        future = Future()
        self.set(future)
        future.set_exception(exception)
        self.last_exception = exception
        for subscriber in self._subscribers:
            subscriber.on_error(exception)

    def close(self):
        future = Future()
        self.set(future)
        future.cancel()

        for subscriber in self._subscribers:
            subscriber.on_completed()

    def ensure_subscribe_now(self):
        self._future.result() if self._future else self._future_set.result()

    def send(self, msg: _T_Change):
        try:
            if self._filter(msg):
                for subscriber in self._subscribers:
                    subscriber.on_next(msg)
        except Exception as e:
            self.error(e)
            return


class ActionObserver:
    def __init__(
        self,
        on_next: Callable[[_T_Change], None],
        on_error: Callable[[Exception], None] = None,
        on_completed: Callable[[], None] = None,
    ):
        self._on_next_callback = on_next
        self._on_error_callback = on_error
        self._on_completed_callback = on_completed

    def on_next(self, value: _T_Change) -> None:
        self._on_next_callback(value)

    def on_error(self, exception: Exception) -> None:
        if self._on_error_callback:
            self._on_error_callback(exception)

    def on_completed(self) -> None:
        if self._on_completed_callback:
            self._on_completed_callback()

    @property
    def on_completed_callback(self):
        return self._on_completed_callback



from __future__ import annotations

import json
from typing import Optional, TYPE_CHECKING, Dict
import requests
from ravendb.http.topology import RaftCommand
from ravendb.documents.operations.definitions import MaintenanceOperation
from ravendb.util.util import RaftIdGenerator
from ravendb.http.raven_command import RavenCommand

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions
    from ravendb.http.server_node import ServerNode


class RefreshConfiguration:
    def __init__(self, disabled: Optional[bool] = None, refresh_frequency_in_sec: Optional[int] = None):
        self.disabled = disabled
        self.refresh_frequency_in_sec = refresh_frequency_in_sec

    def to_json(self) -> Dict:
        return {"Disabled": self.disabled, "RefreshFrequencyInSec": self.refresh_frequency_in_sec}


class ConfigureRefreshOperationResult:
    def __init__(self, raft_command_index: int):
        self.raft_command_index = raft_command_index

    @classmethod
    def from_json(cls, json_dict: Dict) -> ConfigureRefreshOperationResult:
        return cls(json_dict["RaftCommandIndex"])


class ConfigureRefreshOperation(MaintenanceOperation[ConfigureRefreshOperationResult]):
    def __init__(self, configuration: RefreshConfiguration):
        if configuration is None:
            raise ValueError("Configuration cannot be None")
        self._configuration = configuration

    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[ConfigureRefreshOperationResult]":
        return self._ConfigureRefreshCommand(self._configuration)

    class _ConfigureRefreshCommand(RavenCommand[ConfigureRefreshOperationResult], RaftCommand):
        def __init__(self, configuration: RefreshConfiguration):
            super(ConfigureRefreshOperation._ConfigureRefreshCommand, self).__init__(ConfigureRefreshOperationResult)
            self._configuration = configuration

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/admin/refresh/config"
            request = requests.Request("POST", data=self._configuration.to_json())
            request.url = url
            return request

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            self.result = ConfigureRefreshOperationResult.from_json(json.loads(response))

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



from abc import ABC
from typing import Optional, Set, Dict, Callable, Union, TypeVar, List

from ravendb import IndexDefinition, IndexSourceType
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.indexes.definitions import (
    IndexPriority,
    IndexLockMode,
    IndexDeploymentMode,
    IndexState,
    FieldStorage,
    FieldIndexing,
    FieldTermVector,
    IndexFieldOptions,
    IndexType,
)
from ravendb.documents.indexes.abstract_index_creation_tasks import (
    AbstractIndexCreationTaskBase,
    AbstractIndexDefinitionBuilder,
)
from ravendb.documents.indexes.spatial.configuration import SpatialOptions, SpatialOptionsFactory
from ravendb.primitives import constants


_T_IndexDefinition = TypeVar("_T_IndexDefinition", bound=IndexDefinition)


class TimeSeriesIndexDefinition(IndexDefinition):
    @property
    def source_type(self) -> IndexSourceType:
        return IndexSourceType.TIME_SERIES


class TimeSeriesIndexDefinitionBuilder(AbstractIndexDefinitionBuilder[TimeSeriesIndexDefinition]):
    def __init__(self, index_name: Optional[str] = None):
        super().__init__(index_name)
        self.map: Optional[str] = None

    def _new_index_definition(self) -> TimeSeriesIndexDefinition:
        return TimeSeriesIndexDefinition()

    def to_index_definition(
        self, conventions: DocumentConventions, validate_map: bool = True
    ) -> TimeSeriesIndexDefinition:
        if self.map is None and validate_map:
            raise RuntimeError(
                f"Map is required to generate an index, "
                f"you cannot create an index without a valid Map property (in index {self._index_name})."
            )
        return super().to_index_definition(conventions, validate_map)

    def _to_index_definition(
        self, index_definition: TimeSeriesIndexDefinition, conventions: DocumentConventions
    ) -> None:
        if map is None:
            return

        index_definition.maps.add(self.map)


class AbstractGenericTimeSeriesIndexCreationTask(AbstractIndexCreationTaskBase[TimeSeriesIndexDefinition], ABC):
    def __init__(
        self,
        conventions: DocumentConventions = None,
        priority: IndexPriority = None,
        lock_mode: IndexLockMode = None,
        deployment_mode: IndexDeploymentMode = None,
        state: IndexState = None,
    ):
        super().__init__(conventions, priority, lock_mode, deployment_mode, state)
        self._reduce: Optional[str] = None
        self.stores_strings: Dict[str, FieldStorage] = {}
        self.indexes_strings: Dict[str, FieldIndexing] = {}
        self.analyzers_strings: Dict[str, str] = {}
        self.index_suggestions: Set[str] = set()
        self.term_vectors_strings: Dict[str, FieldTermVector] = {}
        self.spatial_options_strings: Dict[str, SpatialOptions] = {}

        self._output_reduce_to_collection: Optional[str] = None
        self._pattern_for_output_reduce_to_collection_references: Optional[str] = None
        self._pattern_references_collection_name: Optional[str] = None

    @property
    def is_map_reduce(self) -> bool:
        return self._reduce is not None

    def _index(self, field: str, indexing: FieldIndexing) -> None:
        self.indexes_strings[field] = indexing

    def _spatial(self, field: str, indexing: Callable[[SpatialOptionsFactory], SpatialOptions]) -> None:
        self.spatial_options_strings[field] = indexing(SpatialOptionsFactory())

    def _store_all_fields(self, storage: FieldStorage) -> None:
        self.stores_strings[constants.Documents.Indexing.Fields.ALL_FIELDS] = storage

    def _store(self, field: str, storage: FieldStorage) -> None:
        self.stores_strings[field] = storage

    def _analyze(self, field: str, analyzer: str) -> None:
        self.analyzers_strings[field] = analyzer

    def _term_vector(self, field: str, term_vector: FieldTermVector) -> None:
        self.term_vectors_strings[field] = term_vector

    def _suggestion(self, field: str) -> None:
        self.index_suggestions.add(field)


class AbstractTimeSeriesIndexCreationTask(AbstractGenericTimeSeriesIndexCreationTask):
    def __init__(
        self,
        conventions: DocumentConventions = None,
        priority: IndexPriority = None,
        lock_mode: IndexLockMode = None,
        deployment_mode: IndexDeploymentMode = None,
        state: IndexState = None,
    ):
        super().__init__(conventions, priority, lock_mode, deployment_mode, state)
        self.map: Optional[str] = None

    def create_index_definition(self) -> Union[IndexDefinition, _T_IndexDefinition]:
        if self.conventions is None:
            self.conventions = DocumentConventions()

        index_definition_builder = TimeSeriesIndexDefinitionBuilder(self.index_name)
        index_definition_builder.indexes_strings = self.indexes_strings
        index_definition_builder.analyzers_strings = self.analyzers_strings
        index_definition_builder.map = self.map
        index_definition_builder.reduce = self._reduce
        index_definition_builder.stores_strings = self.stores_strings
        index_definition_builder.suggestions_options = self.index_suggestions
        index_definition_builder.term_vector_strings = self.term_vectors_strings
        index_definition_builder.spatial_indexes_strings = self.spatial_options_strings
        index_definition_builder.output_reduce_to_collection = self._output_reduce_to_collection
        index_definition_builder.pattern_for_output_reduce_to_collection_references = (
            self._pattern_for_output_reduce_to_collection_references
        )
        index_definition_builder.pattern_references_collection_name = self._pattern_references_collection_name
        index_definition_builder.additional_sources = self.additional_sources
        index_definition_builder.additional_assemblies = self.additional_assemblies
        index_definition_builder.configuration = self.configuration
        index_definition_builder.lock_mode = self.lock_mode
        index_definition_builder.priority = self.priority
        index_definition_builder.state = self.state
        index_definition_builder.deployment_mode = self.deployment_mode

        return index_definition_builder.to_index_definition(self.conventions)


class AbstractMultiMapTimeSeriesIndexCreationTask(AbstractGenericTimeSeriesIndexCreationTask):
    def __init__(
        self,
        conventions: DocumentConventions = None,
        priority: IndexPriority = None,
        lock_mode: IndexLockMode = None,
        deployment_mode: IndexDeploymentMode = None,
        state: IndexState = None,
    ):
        super().__init__(conventions, priority, lock_mode, deployment_mode, state)
        self.maps: List[str] = []

    def _add_map(self, map: str) -> None:
        if map is None:
            raise ValueError("Map cannot be None.")
        self.maps.append(map)

    def create_index_definition(self) -> Union[IndexDefinition, _T_IndexDefinition]:
        if self.conventions is None:
            self.conventions = DocumentConventions()

        index_definition_builder = TimeSeriesIndexDefinitionBuilder(self.index_name)
        index_definition_builder.indexes_strings = self.indexes_strings
        index_definition_builder.analyzers_strings = self.analyzers_strings
        index_definition_builder.reduce = self._reduce
        index_definition_builder.stores_strings = self.stores_strings
        index_definition_builder.suggestions_options = self.index_suggestions
        index_definition_builder.term_vector_strings = self.term_vectors_strings
        index_definition_builder.spatial_indexes_strings = self.spatial_options_strings
        index_definition_builder.output_reduce_to_collection = self._output_reduce_to_collection
        index_definition_builder.pattern_for_output_reduce_to_collection_references = (
            self._pattern_for_output_reduce_to_collection_references
        )
        index_definition_builder.pattern_references_collection_name = self._pattern_references_collection_name
        index_definition_builder.additional_sources = self.additional_sources
        index_definition_builder.additional_assemblies = self.additional_assemblies
        index_definition_builder.configuration = self.configuration
        index_definition_builder.lock_mode = self.lock_mode
        index_definition_builder.priority = self.priority
        index_definition_builder.state = self.state
        index_definition_builder.deployment_mode = self.deployment_mode

        index_definition = index_definition_builder.to_index_definition(self.conventions, False)
        index_definition.maps = set(self.maps)

        return index_definition


class AbstractJavaScriptTimeSeriesIndexCreationTask(AbstractIndexCreationTaskBase[TimeSeriesIndexDefinition]):
    def __init__(
        self,
        conventions: DocumentConventions = None,
        priority: IndexPriority = None,
        lock_mode: IndexLockMode = None,
        deployment_mode: IndexDeploymentMode = None,
        state: IndexState = None,
    ):
        super().__init__(conventions, priority, lock_mode, deployment_mode, state)
        self._definition = TimeSeriesIndexDefinition()

    @property
    def maps(self) -> Set[str]:
        return self._definition.maps

    @maps.setter
    def maps(self, maps: Set[str]):
        self._definition.maps = maps

    @property
    def fields(self) -> Dict[str, IndexFieldOptions]:
        return self._definition.fields

    @fields.setter
    def fields(self, fields: Dict[str, IndexFieldOptions]):
        self._definition.fields = fields

    @property
    def reduce(self) -> str:
        return self._definition.reduce

    @reduce.setter
    def reduce(self, reduce: str):
        self._definition.reduce = reduce

    def is_map_reduce(self) -> bool:
        return self.reduce is not None

    @property
    def output_reduce_to_collection(self) -> str:
        return self._definition.output_reduce_to_collection

    @output_reduce_to_collection.setter
    def output_reduce_to_collection(self, output_reduce_to_collection: str):
        self._definition.output_reduce_to_collection = output_reduce_to_collection

    @property
    def pattern_references_collection_name(self) -> str:
        return self._definition.pattern_references_collection_name

    @pattern_references_collection_name.setter
    def pattern_references_collection_name(self, pattern_references_collection_name: str):
        self._definition.pattern_references_collection_name = pattern_references_collection_name

    @property
    def pattern_for_output_reduce_to_collection_references(self) -> str:
        return self._definition.pattern_for_output_reduce_to_collection_references

    @pattern_for_output_reduce_to_collection_references.setter
    def pattern_for_output_reduce_to_collection_references(self, pattern_for_output_reduce_to_collection_references):
        self._definition.pattern_for_output_reduce_to_collection_references = (
            pattern_for_output_reduce_to_collection_references
        )

    def create_index_definition(self) -> TimeSeriesIndexDefinition:
        self._definition.name = self.index_name
        self._definition.type = IndexType.JAVA_SCRIPT_MAP_REDUCE if self.is_map_reduce() else IndexType.JAVA_SCRIPT_MAP
        self._definition.additional_sources = self.additional_sources or {}
        self._definition.additional_assemblies = self.additional_assemblies or set()
        self._definition.configuration = self.configuration
        self._definition.lock_mode = self.lock_mode
        self._definition.priority = self.priority
        self._definition.state = self.state
        self._definition.deployment_mode = self.deployment_mode
        return self._definition



from __future__ import annotations

import inspect
import threading
from abc import abstractmethod, ABC
from datetime import timedelta, datetime
from enum import Enum
from typing import Dict, List, Tuple, Callable, Union, Optional, Generic, Type, Any, TYPE_CHECKING

import inflect

from typing import TypeVar
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
from ravendb.primitives import constants
from ravendb.documents.operations.configuration.definitions import (
    ClientConfiguration,
    LoadBalanceBehavior,
    ReadBalanceBehavior,
)
from ravendb.documents.indexes.definitions import SortOptions
from ravendb.tools.utils import Utils

inflect.def_classical["names"] = False
inflector = inflect.engine()

_T = TypeVar("_T")

if TYPE_CHECKING:
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )


class DocumentConventions(object):
    @classmethod
    def default_conventions(cls):
        return cls()

    __cached_default_type_collection_names: Dict[type, str] = {}
    __cached_keys_collection_names: Dict[str, str] = {}

    def __init__(self):
        self._frozen = False

        # Value constraints
        self.identity_parts_separator = "/"
        self.max_number_of_requests_per_session = 30
        self._max_http_cache_size = 128 * 1024 * 1024
        self.max_length_of_query_using_get_url = 1024 + 512
        self.time_series_batch_size = 1024

        # Flags
        self.disable_topology_updates = False
        self.use_optimistic_concurrency = False
        self.throw_if_query_page_size_is_not_set = False
        self._send_application_identifier = True
        self._save_enums_as_integers: Optional[bool] = None
        self._disable_atomic_document_writes_in_cluster_wide_transaction: Optional[bool] = None

        # Configuration
        self.json_default_method = DocumentConventions.json_default
        self._original_configuration: Optional[ClientConfiguration] = None
        self._should_ignore_entity_changes: Optional[ShouldIgnoreEntityChanges] = None

        # Collections
        self._list_of_registered_id_conventions: List[Tuple[Type, Callable[[str, object], str]]] = []
        self._list_of_query_value_to_object_converters: List[Tuple[Type, ValueForQueryConverter[object]]] = []

        # Utilities
        self.document_id_generator: Optional[Callable[[str, object], str]] = None
        self._find_identity_property_name: Callable[[Type[Any]], str] = lambda type_: "Id"
        self._id_property_name_cache: Dict[Type, str] = {}
        self._find_python_class: Optional[Callable[[str, Dict], str]] = None
        self._find_collection_name: Callable[[Type], str] = self.default_get_collection_name
        self._find_collection_name_for_dict: Callable[[str], str] = self.default_get_collection_name_for_dict
        self._find_python_class_name: Callable[[Type], str] = (
            lambda object_type: f"{object_type.__module__}.{object_type.__name__}"
        )
        self._transform_class_collection_name_to_document_id_prefix = (
            lambda collection_name: self.default_transform_collection_name_to_document_id_prefix(collection_name)
        )

        # Timeouts
        self.request_timeout: timedelta = timedelta.min
        self.first_broadcast_attempt_timeout = timedelta(seconds=5)
        self.second_broadcast_attempt_timeout = timedelta(seconds=30)
        self.wait_for_indexes_after_save_changes_timeout = timedelta(seconds=15)
        self.wait_for_replication_after_save_changes_timeout = timedelta(seconds=15)
        self.wait_for_non_stale_results_timeout = timedelta(seconds=15)

        # Balancing
        self._load_balancer_context_seed: Optional[int] = None
        self._load_balance_behavior: Optional[LoadBalanceBehavior] = LoadBalanceBehavior.NONE
        self._read_balance_behavior: Optional[ReadBalanceBehavior] = ReadBalanceBehavior.NONE
        self._load_balancer_per_session_context_selector: Optional[Callable[[str], str]] = None

        # Async
        self._update_from_lock = threading.Lock()

    def freeze(self):
        self._frozen = True

    def is_frozen(self):
        return self._frozen

    def get_python_class_name(self, entity_type: type):
        return self._find_python_class_name(entity_type)

    @property
    def find_collection_name(self) -> Callable[[type], str]:
        return self._find_collection_name

    @find_collection_name.setter
    def find_collection_name(self, value) -> None:
        self._assert_not_frozen()
        self._find_collection_name = value

    @property
    def find_python_class(self) -> Callable[[str, Dict], Optional[str]]:
        def __default(key: str, doc: Dict) -> Optional[str]:
            metadata = doc.get(constants.Documents.Metadata.KEY)
            if metadata:
                python_type = metadata.get(constants.Documents.Metadata.RAVEN_PYTHON_TYPE)
                return python_type
            return None

        return self._find_python_class or __default

    @find_python_class.setter
    def find_python_class(self, value: Callable[[str, Dict], str]):
        self._assert_not_frozen()
        self._find_python_class = value

    def get_python_class(self, key: str, document: Dict) -> str:
        return self.find_python_class(key, document)

    @property
    def transform_class_collection_name_to_document_id_prefix(self) -> Callable[[str], str]:
        return self._transform_class_collection_name_to_document_id_prefix

    @transform_class_collection_name_to_document_id_prefix.setter
    def transform_class_collection_name_to_document_id_prefix(self, value: Callable[[str], str]) -> None:
        self._assert_not_frozen()
        self._transform_class_collection_name_to_document_id_prefix = value

    @property
    def load_balancer_per_session_context_selector(self) -> Callable[[str], str]:
        return self._load_balancer_per_session_context_selector

    @load_balancer_per_session_context_selector.setter
    def load_balancer_per_session_context_selector(self, value: Callable[[str], str]):
        self.load_balancer_per_session_context_selector = value

    @property
    def max_http_cache_size(self) -> int:
        return self._max_http_cache_size

    @max_http_cache_size.setter
    def max_http_cache_size(self, value: int):
        self._max_http_cache_size = value

    @property
    def save_enums_as_integers(self) -> bool:
        return self._save_enums_as_integers

    @save_enums_as_integers.setter
    def save_enums_as_integers(self, value: bool):
        self._save_enums_as_integers = value

    @property
    def find_python_class_name(self) -> Callable[[type], str]:
        return self._find_python_class_name

    @find_python_class_name.setter
    def find_python_class_name(self, value) -> None:
        self._assert_not_frozen()
        self._find_python_class_name = value

    @property
    def should_ignore_entity_changes(self) -> ShouldIgnoreEntityChanges:
        return self._should_ignore_entity_changes

    @should_ignore_entity_changes.setter
    def should_ignore_entity_changes(self, value: ShouldIgnoreEntityChanges) -> None:
        self._assert_not_frozen()
        self._should_ignore_entity_changes = value

    @property
    def load_balancer_context_seed(self) -> int:
        return self._load_balancer_context_seed

    @load_balancer_context_seed.setter
    def load_balancer_context_seed(self, value: int):
        self._assert_not_frozen()
        self._load_balancer_context_seed = value

    @property
    def load_balance_behavior(self):
        return self._load_balance_behavior

    @load_balance_behavior.setter
    def load_balance_behavior(self, value: LoadBalanceBehavior):
        self._assert_not_frozen()
        self._load_balance_behavior = value

    @property
    def read_balance_behavior(self) -> ReadBalanceBehavior:
        return self._read_balance_behavior

    @read_balance_behavior.setter
    def read_balance_behavior(self, value: ReadBalanceBehavior):
        self._assert_not_frozen()
        self._read_balance_behavior = value

    @property
    def send_application_identifier(self) -> bool:
        return self._send_application_identifier

    @property
    def disable_atomic_document_writes_in_cluster_wide_transaction(self) -> bool:
        return self._disable_atomic_document_writes_in_cluster_wide_transaction

    @disable_atomic_document_writes_in_cluster_wide_transaction.setter
    def disable_atomic_document_writes_in_cluster_wide_transaction(self, value: bool):
        self._assert_not_frozen()
        self._disable_atomic_document_writes_in_cluster_wide_transaction = value

    @property
    def find_identity_property_name(self) -> Callable[[Type[Any]], str]:
        return self._find_identity_property_name

    @find_identity_property_name.setter
    def find_identity_property_name(self, find_identity_property_name_function: Callable[[Type[Any]], str]):
        self._find_identity_property_name = find_identity_property_name_function

    @staticmethod
    def json_default(o):
        if o is None:
            return None
        if isinstance(o, datetime):
            return Utils.datetime_to_string(o)
        elif isinstance(o, timedelta):
            return Utils.timedelta_to_str(o)
        elif isinstance(o, Enum):
            return o.value
        elif isinstance(o, MetadataAsDictionary):
            return o.metadata
        elif getattr(o, "to_json", None) and getattr(o.to_json, "__call__", None):
            return o.to_json()
        elif getattr(o, "__dict__", None):
            return o.__dict__
        elif isinstance(o, set):
            return list(o)
        elif isinstance(o, (int, float)):
            return str(o)
        else:
            raise TypeError(
                repr(o) + " is not JSON serializable (Try add a json default method to convention"
                " or try to add methods - to_json & classmethod from_json - to object class)"
            )

    @staticmethod
    def default_transform_plural(name):
        return inflector.plural(name)

    @staticmethod
    def default_transform_type_tag_name(name):
        count = sum(1 for c in name if c.isupper())
        if count <= 1:
            return DocumentConventions.default_transform_plural(name.lower())
        return DocumentConventions.default_transform_plural(name)

    @staticmethod
    def build_default_metadata(entity):
        if entity is None:
            return {}
        existing = entity.__dict__.get("@metadata")
        if existing is None:
            existing = {}

        new_metadata = {
            "@collection": DocumentConventions.default_transform_plural(entity.__class__.__name__),
            "Raven-Python-Type": "{0}.{1}".format(entity.__class__.__module__, entity.__class__.__name__),
        }

        existing.update(new_metadata)

        return existing

    def get_collection_name(self, entity_or_type: Union[type, object]) -> str:
        if not entity_or_type:
            return None
        object_type = type(entity_or_type) if not isinstance(entity_or_type, type) else entity_or_type
        collection_name = self._find_collection_name(object_type)
        if collection_name:
            return collection_name

        return self.default_get_collection_name(object_type)

    def get_collection_name_for_dict(self, key: str):
        collection = key.split("/")[0]
        collection_name = self._find_collection_name_for_dict(collection)
        if collection_name:
            return collection_name
        return self.default_get_collection_name(dict)

    def generate_document_id(self, database_name: str, entity: object) -> str:
        object_type = type(entity)
        for list_of_registered_id_convention in self._list_of_registered_id_conventions:
            return list_of_registered_id_convention[1](database_name, entity)
        return self.document_id_generator(database_name, entity)

    @staticmethod
    def default_get_collection_name(object_type: type) -> str:
        result = DocumentConventions.__cached_default_type_collection_names.get(object_type)
        if result:
            return result
        # we want to reject queries and other operations on abstract types, because you usually
        # want to use them for polymorphic queries, and that require the conventions to be
        # applied properly, so we reject the behavior and hint to the user explicitly
        if inspect.isabstract(object_type):
            raise ValueError(
                f"Cannot find collection name for abstract class {object_type}, "
                f"only concrete class are supported. "
                f"Did you forget to customize conventions.find_collection_name?"
            )
        result = inflector.plural(str(object_type.__name__))  # todo: hilo multidb problems
        DocumentConventions.__cached_default_type_collection_names[object_type] = result
        return result

    @staticmethod
    def default_get_collection_name_for_dict(key: str) -> str:
        result = DocumentConventions.__cached_keys_collection_names.get(key, None)
        if result:
            return result
        # singular_noun returns False if the word is singular
        result = inflector.plural(key) if not inflector.singular_noun(key) else key
        DocumentConventions.__cached_keys_collection_names[key] = result
        return result

    @staticmethod
    def try_get_type_from_metadata(metadata: Dict[str, Any]) -> Optional[str]:
        if "Raven-Python-Type" in metadata:
            return metadata["Raven-Python-Type"]
        return None

    @staticmethod
    def uses_range_type(obj):
        if obj is None:
            return False

        if isinstance(obj, int) or isinstance(obj, float):
            return True
        return False

    @staticmethod
    def get_default_sort_option(type_name):
        if not type_name:
            return None
        if type_name == "int" or type_name == "float" or type_name == "long":
            return SortOptions.numeric

    @staticmethod
    def range_field_name(field_name, type_name):
        if type_name == "long":
            field_name = "{0}_L_Range".format(field_name)
        else:
            field_name = "{0}_D_Range".format(field_name)

        return field_name

    def _assert_not_frozen(self) -> None:
        if self._frozen:
            raise RuntimeError(
                "Conventions has been frozen after documentStore.initialize()" " and no changes can be applied to them"
            )

    def clone(self) -> DocumentConventions:
        cloned = DocumentConventions()
        cloned._list_of_registered_id_conventions = [*self._list_of_registered_id_conventions]
        cloned._frozen = self._frozen
        cloned._should_ignore_entity_changes = self._should_ignore_entity_changes
        cloned._original_configuration = self._original_configuration
        cloned._save_enums_as_integers = self._save_enums_as_integers
        cloned.identity_parts_separator = self.identity_parts_separator
        cloned.disable_topology_updates = self.disable_topology_updates
        cloned._find_identity_property_name = self._find_identity_property_name

        cloned.document_id_generator = self.document_id_generator

        cloned._find_collection_name = self._find_collection_name
        cloned._find_python_class_name = self.find_python_class_name

        cloned.use_optimistic_concurrency = self.use_optimistic_concurrency
        cloned.throw_if_query_page_size_is_not_set = self.throw_if_query_page_size_is_not_set
        cloned.max_number_of_requests_per_session = self.max_number_of_requests_per_session

        cloned._read_balance_behavior = self._read_balance_behavior
        cloned._load_balance_behavior = self._load_balance_behavior
        self._max_http_cache_size = self._max_http_cache_size
        return cloned

    def get_identity_property_name(self, object_type: Type[Any]) -> Optional[str]:
        # Check the cache first
        if object_type in self._id_property_name_cache:
            return self._id_property_name_cache[object_type]

        id_property_name = self.find_identity_property_name(object_type)

        # Cache the result
        self._id_property_name_cache[object_type] = id_property_name

        return id_property_name

    def update_from(self, configuration: ClientConfiguration):
        if configuration.disabled and self._original_configuration is None:
            return
        with self._update_from_lock:
            if configuration.disabled and self._original_configuration is not None:
                self.max_number_of_requests_per_session = (
                    self._original_configuration.max_number_of_requests_per_session
                    if self._original_configuration.max_number_of_requests_per_session
                    else self.max_number_of_requests_per_session
                )
                self._read_balance_behavior = (
                    self._original_configuration.read_balance_behavior
                    if self._original_configuration.read_balance_behavior
                    else self._read_balance_behavior
                )
                self.identity_parts_separator = (
                    self._original_configuration.identity_parts_separator
                    if self._original_configuration.identity_parts_separator
                    else self.identity_parts_separator
                )
                self._load_balance_behavior = (
                    self._original_configuration.load_balance_behavior
                    if self._original_configuration.load_balance_behavior
                    else self._load_balance_behavior
                )
                self._load_balancer_context_seed = (
                    self._original_configuration.load_balancer_context_seed
                    if self._original_configuration.load_balancer_context_seed
                    else self._load_balancer_context_seed
                )

                self._original_configuration = None
                return

            if self._original_configuration is None:
                self._original_configuration = ClientConfiguration()
                self._original_configuration.etag = -1
                self._original_configuration.max_number_of_requests_per_session = (
                    self.max_number_of_requests_per_session
                )
                self._original_configuration.__read_balance_behavior = self._read_balance_behavior
                self._original_configuration.identity_parts_separator = self.identity_parts_separator
                self._original_configuration.__load_balance_behavior = self._load_balance_behavior
                self._original_configuration.__load_balancer_context_seed = self._load_balancer_context_seed

            # first not None
            self.max_number_of_requests_per_session = next(
                item
                for item in [
                    configuration.max_number_of_requests_per_session,
                    self._original_configuration.max_number_of_requests_per_session,
                    self.max_number_of_requests_per_session,
                ]
                if item is not None
            )

            self._read_balance_behavior = next(
                item
                for item in [
                    configuration.read_balance_behavior,
                    self._original_configuration.read_balance_behavior,
                    self._read_balance_behavior,
                ]
                if item is not None
            )

            self.identity_parts_separator = next(
                item
                for item in [
                    configuration.identity_parts_separator,
                    self._original_configuration.identity_parts_separator,
                    self.identity_parts_separator,
                ]
                if item is not None
            )

            self._load_balance_behavior = next(
                item
                for item in [
                    configuration.load_balance_behavior,
                    self._original_configuration.load_balance_behavior,
                    self._load_balance_behavior,
                ]
                if item is not None
            )

            self._load_balancer_context_seed = (
                configuration.load_balancer_context_seed
                or self._original_configuration.load_balancer_context_seed
                or self._load_balancer_context_seed
            )

    @staticmethod
    def default_transform_collection_name_to_document_id_prefix(collection_name: str) -> str:
        upper_count = len(list(filter(str.isupper, [char for char in collection_name])))
        if upper_count <= 1:
            return collection_name.lower()

        # multiple capital letters, so probably something that we want to preserve caps on.
        return collection_name

    def try_convert_value_to_object_for_query(
        self, field_name: str, value: object, for_range: bool
    ) -> Tuple[bool, object]:
        for query_value_converter in self._list_of_query_value_to_object_converters:
            if not isinstance(value, query_value_converter[0]):
                continue

            return query_value_converter[1].try_to_convert_value_for_query(field_name, value, for_range)

        return False, None


class ValueForQueryConverter(Generic[_T]):
    @abstractmethod
    def try_to_convert_value_for_query(self, field_name: str, value: _T, for_range: bool) -> Tuple[bool, object]:
        pass


class ShouldIgnoreEntityChanges(ABC):
    @abstractmethod
    def check(
        self,
        session_operations: "InMemoryDocumentSessionOperations",
        entity: object,
        document_id: str,
    ) -> bool:
        pass



from typing import TYPE_CHECKING, Optional

import requests

from ravendb.documents.session.misc import DocumentQueryCustomization
from ravendb.http.topology import Topology

if TYPE_CHECKING:
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )


class EventArgs:
    pass  # empty by design


class BeforeStoreEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations", document_id: str, entity: object):
        self.__session = session
        self.__document_id = document_id
        self.__entity = entity
        self.__document_metadata = None

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.__session

    @property
    def document_id(self) -> str:
        return self.__document_id

    @property
    def entity(self) -> object:
        return self.__entity

    @property
    def is_metadata_accessed(self) -> bool:
        return self.__document_metadata is not None

    @property
    def document_metadata(self):
        if self.__document_metadata is None:
            self.__document_metadata = self.__session.advanced.get_metadata_for(self.__entity)
        return self.__document_metadata


class SessionClosingEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations"):
        self.__session = session

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.__session


class BeforeConversionToDocumentEventArgs(EventArgs):
    def __init__(self, key: str, entity: object, session: "InMemoryDocumentSessionOperations"):
        self.__key = key
        self.__entity = entity
        self.__session = session

    @property
    def key(self) -> str:
        return self.__key

    @property
    def entity(self) -> object:
        return self.__entity

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.__session


class AfterConversionToDocumentEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations", key: str, entity: object, document: dict):
        self.__session = session
        self.__key = key
        self.__entity = entity
        self.__document = document

    @property
    def key(self) -> str:
        return self.__key

    @property
    def entity(self) -> object:
        return self.__entity

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.__session

    @property
    def document(self) -> dict:
        return self.__document


class BeforeConversionToEntityEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations", key: str, object_type: type, document: dict):
        self.__session = session
        self.__key = key
        self.__object_type = object_type
        self.__document = document

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.__session

    @property
    def key(self) -> str:
        return self.__key

    @property
    def object_type(self) -> type:
        return self.__object_type

    @property
    def document(self) -> dict:
        return self.__document


class AfterConversionToEntityEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations", key: str, document: dict, entity: object):
        self.__session = session
        self.__key = key
        self.__document = document
        self.__entity = entity

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.__session

    @property
    def key(self) -> str:
        return self.__key

    @property
    def document(self) -> dict:
        return self.__document

    @property
    def entity(self) -> object:
        return self.__entity


class AfterSaveChangesEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations", document_id: str, entity: object):
        self.__session = session
        self.__document_id = document_id
        self.__entity = entity

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.__session

    @property
    def document_id(self) -> str:
        return self.__document_id

    @property
    def entity(self) -> object:
        return self.__entity


class BeforeDeleteEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations", key: str, entity: object):
        self.__session = session
        self.__key = key
        self.__entity = entity

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.session

    @property
    def key(self) -> str:
        return self.__key

    @property
    def entity(self) -> object:
        return self.__entity


class BeforeQueryEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations", query_customization: DocumentQueryCustomization):
        self.__session = session
        self.__query_customization = query_customization

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.session

    @property
    def query_customization(self) -> DocumentQueryCustomization:
        return self.__query_customization


class SessionCreatedEventArgs(EventArgs):
    def __init__(self, session: "InMemoryDocumentSessionOperations"):
        self.__session = session

    @property
    def session(self) -> "InMemoryDocumentSessionOperations":
        return self.__session


class BeforeRequestEventArgs(EventArgs):
    def __init__(self, database: str, url: str, request: requests.Request, attempt_number: int):
        self.__database = database
        self.__url = url
        self.__request = request
        self.__attempt_number = attempt_number

    @property
    def database(self) -> str:
        return self.__database

    @property
    def url(self) -> str:
        return self.__url

    @property
    def request(self) -> requests.Request:
        return self.__request

    @property
    def attempt_number(self) -> int:
        return self.__attempt_number


class SucceedRequestEventArgs(EventArgs):
    def __init__(
        self, database: str, url: str, response: requests.Response, request: requests.Request, attempt_number: int
    ):
        self.__database = database
        self.__url = url
        self.__response = response
        self.__request = request
        self.__attempt_number = attempt_number

    @property
    def database(self) -> str:
        return self.__database

    @property
    def url(self) -> str:
        return self.__url

    @property
    def response(self) -> requests.Response:
        return self.__response

    @property
    def request(self) -> requests.Request:
        return self.__request

    @property
    def attempt_number(self) -> int:
        return self.__attempt_number


class FailedRequestEventArgs(EventArgs):
    def __init__(
        self,
        database: str,
        url: str,
        exception: Exception,
        request: Optional[requests.Request],
        response: Optional[requests.Response],
    ):
        self.__database = database
        self.__url = url
        self.__exception = exception
        self.__request = request
        self.__response = response

    @property
    def database(self) -> str:
        return self.__database

    @property
    def url(self) -> str:
        return self.__url

    @property
    def exception(self) -> Exception:
        return self.__exception

    @property
    def request(self) -> requests.Request:
        return self.__request

    @property
    def response(self) -> requests.Response:
        return self.__response


class TopologyUpdatedEventArgs(EventArgs):
    def __init__(self, topology: Topology):
        self.__topology = topology

    @property
    def topology(self) -> Topology:
        return self.__topology



from __future__ import annotations
from abc import ABC
from datetime import datetime
from enum import Enum
from typing import Dict

from ravendb.tools.utils import Utils


class DatabaseChange(ABC):
    pass


class DocumentChangeType(Enum):
    NONE = "None"
    PUT = "Put"
    DELETE = "Delete"
    CONFLICT = "Conflict"
    COMMON = "Common"

    def __str__(self):
        return self.value


class DocumentChange(DatabaseChange):
    def __init__(self, type_of_change: DocumentChangeType, key: str, collection_name: str, change_vector: str):
        self.type_of_change = type_of_change
        self.key = key
        self.collection_name = collection_name
        self.change_vector = change_vector

    def __str__(self):
        return self.type_of_change.value + " on " + self.key

    @classmethod
    def from_json(cls, json_dict: Dict) -> DocumentChange:
        return cls(
            DocumentChangeType(json_dict["Type"]),
            json_dict["Id"],
            json_dict["CollectionName"],
            json_dict["ChangeVector"],
        )


class IndexChangeTypes(Enum):
    NONE = "None"

    BATCH_COMPLETED = "BatchCompleted"

    INDEX_ADDED = "IndexAdded"
    INDEX_REMOVED = "IndexRemoved"

    INDEX_DEMOTED_TO_IDLE = "IndexDemotedToIdle"
    INDEX_PROMOTED_FROM_IDLE = "IndexPromotedFromIdle"

    INDEX_DEMOTED_TO_DISABLED = "IndexDemotedToDisabled"

    INDEX_MARKED_AS_ERRORED = "IndexMarkedAsErrored"

    SIDE_BY_SIDE_REPLACE = "SideBySideReplace"

    RENAMED = "Renamed"
    INDEX_PAUSED = "IndexPaused"
    LOCK_MODE_CHANGED = "LockModeChanged"
    PRIORITY_CHANGED = "PriorityChanged"

    ROLLING_INDEX_CHANGED = "RollingIndexChanged"

    def __str__(self):
        return self.value


class IndexChange(DatabaseChange):
    def __init__(self, type_of_change: IndexChangeTypes, name: str):
        self.type_of_change = type_of_change
        self.name = name

    @classmethod
    def from_json(cls, json_dict: Dict) -> IndexChange:
        return cls(IndexChangeTypes(json_dict["Type"]), json_dict["Name"])


class TimeSeriesChangeTypes(Enum):
    NONE = "None"
    PUT = "Put"
    DELETE = "Delete"
    MIXED = "Mixed"

    def __str__(self):
        return self.value


class TimeSeriesChange(DatabaseChange):
    def __init__(
        self,
        name: str,
        from_: datetime,
        to_: datetime,
        document_id: str,
        change_vector: str,
        type_of_change: TimeSeriesChangeTypes,
        collection_name: str,
    ):
        self.name = name
        self.from_ = from_
        self.to_ = to_
        self.document_id = document_id
        self.change_vector = change_vector
        self.type_of_change = type_of_change
        self.collection_name = collection_name

    @classmethod
    def from_json(cls, json_dict: Dict) -> TimeSeriesChange:
        return cls(
            json_dict["Name"],
            Utils.string_to_datetime(json_dict["From"]),
            Utils.string_to_datetime(json_dict["To"]),
            json_dict["DocumentId"],
            json_dict["ChangeVector"],
            TimeSeriesChangeTypes(json_dict["Type"]),
            json_dict["CollectionName"],
        )


class CounterChangeTypes(Enum):
    NONE = "None"
    PUT = "Put"
    DELETE = "Delete"
    INCREMENT = "Increment"

    def __str__(self):
        return self.value


class CounterChange(DatabaseChange):
    def __init__(
        self,
        name: str,
        value: int,
        document_id: str,
        collection_name: str,
        change_vector: str,
        type_of_change: CounterChangeTypes,
    ):
        self.name = name
        self.value = value
        self.document_id = document_id
        self.collection_name = collection_name
        self.change_vector = change_vector
        self.type_of_change = type_of_change

    @classmethod
    def from_json(cls, json_dict: Dict) -> CounterChange:
        return cls(
            json_dict["Name"],
            json_dict["Value"],
            json_dict["DocumentId"],
            json_dict["CollectionName"],
            json_dict["ChangeVector"],
            CounterChangeTypes(json_dict["Type"]),
        )


class OperationStatusChange(DatabaseChange):
    def __init__(self, operation_id: str, state: Dict):
        self.operation_id = operation_id
        self.state = state

    @classmethod
    def from_json(cls, json_dict: Dict) -> OperationStatusChange:
        return OperationStatusChange(json_dict["OperationId"], json_dict["State"])


class TopologyChange(DatabaseChange):
    def __init__(self, url: str, database: str):
        self.url = url
        self.database = database

    @classmethod
    def from_json(cls, json_dict: Dict) -> TopologyChange:
        return cls(json_dict["Url"], json_dict["Database"])



from __future__ import annotations

import http
import os

from ravendb.exceptions.documents import DocumentConflictException
from ravendb.exceptions.raven_exceptions import ConcurrencyException, RavenException


class ExceptionDispatcher:
    class ExceptionSchema:
        def __init__(self, url: str = None, object_type: str = None, message: str = None, error: str = None):
            self.url = url
            self.type = object_type
            self.message = message
            self.error = error

    @staticmethod
    def get(schema: ExceptionDispatcher.ExceptionSchema, code: int, inner: Exception = None) -> RavenException:
        message = schema.message
        type_as_string = schema.type

        if code == http.HTTPStatus.CONFLICT:
            if "DocumentConflictException" in type_as_string:
                return DocumentConflictException.from_message(message)
            return ConcurrencyException(message)

        error = f"{schema.error}{os.linesep}The server at {schema.url} responded with status code: {code}"

        error_type = ExceptionDispatcher.__get_type(type_as_string)
        if error_type is None:
            return RavenException(error, inner)

        try:
            exception = error_type(error)
        except BaseException as e:
            return RavenException(error, inner)

        if not issubclass(error_type, RavenException):
            return RavenException(error, exception)

        return exception

    @staticmethod
    def __get_type(type_as_string: str) -> type:
        if "System.TimeoutException" == type_as_string:
            return TimeoutError

        prefix = "Raven.Client.Exceptions."
        if type_as_string.startswith(prefix):
            exception_name = type_as_string[len(prefix) : :]
            if "." in exception_name:
                exception_name = ".".join(list(map(str.lower, exception_name.split("."))))

            try:
                return __import__(f"pyravendb.exceptions.{exception_name}")  # todo: fix, doesn't work
            except Exception as e:
                return None
        else:
            return None



from __future__ import annotations
from abc import abstractmethod
from typing import TypeVar, TYPE_CHECKING, Generic, Dict, Any

from ravendb.http.http_cache import HttpCache

_T = TypeVar("_T")
_Operation_T = TypeVar("_Operation_T")

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions
    from ravendb.documents.store.definition import DocumentStore
    from ravendb.http.raven_command import RavenCommand, VoidRavenCommand


class IOperation(Generic[_Operation_T]):
    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: HttpCache
    ) -> "RavenCommand[_Operation_T]":
        pass


class VoidOperation(IOperation[None]):
    @abstractmethod
    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: HttpCache
    ) -> "VoidRavenCommand":
        pass


class MaintenanceOperation(Generic[_T]):
    @abstractmethod
    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[_T]":
        pass


class VoidMaintenanceOperation(MaintenanceOperation[None]):
    @abstractmethod
    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        pass


class OperationIdResult:
    def __init__(self, operation_id: int = None, operation_node_tag: str = None):
        self.operation_id = operation_id
        self.operation_node_tag = operation_node_tag

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> "OperationIdResult":
        return cls(json_dict["OperationId"], json_dict["OperationNodeTag"])


class OperationExceptionResult:
    def __init__(self, type: str, message: str, error: str, status_code: int):
        self.type = type
        self.message = message
        self.error = error
        self.status_code = status_code

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> OperationExceptionResult:
        return cls(json_dict["Type"], json_dict["Message"], json_dict["Error"], json_dict["StatusCode"])



from ravendb.tools.utils import Utils
import json


class JsonDecoder(json.JSONDecoder):
    """
    This custom JsonDecoder can add to json.loads function to work with ravendb mappers solution.
    just use it like this json.loads(YOUR_OBJECT, cls=JsonDecoder, object_mapper=YOUR_MAPPER)
    Note that that last object that returns from the loads function will be a dict.
    To get the dict as your custom object you can create it with the return dict or use the parse_json method
    """

    def __init__(self, **kwargs):
        self.object_mapper = kwargs.pop("object_mapper", lambda key, value: None)
        super(JsonDecoder, self).__init__(object_hook=self.object_hook)
        self.step_down = None

    def object_hook(self, dict_object):
        if self.step_down is not None:
            for key in dict_object:
                result = self.object_mapper(key, dict_object[key])
                if result is not None:
                    dict_object[key] = result
        self.step_down = dict_object
        return self.step_down


def parse_json(json_string, object_type, mappers, convert_to_snake_case=False):
    """
    This function will use the custom JsonDecoder and the conventions.mappers to recreate your custom object
    in the parse json string state just call this method with the json_string your complete object_type and with your
    mappers dict.
    the mappers dict must contain in the key the object_type (ex. User) and the value will contain a method that get
    key, value (the key will be the name of the object property we like to parse and the value
    will be the properties of the object)
    """
    obj = json.loads(json_string, cls=JsonDecoder, object_mapper=mappers.get(object_type, None))

    if obj is not None:
        try:
            obj = object_type(**obj)
        except TypeError:
            initialize_dict, set_needed = Utils.make_initialize_dict(obj, object_type.__init__, convert_to_snake_case)
            o = object_type(**initialize_dict)
            if set_needed:
                for key, value in obj.items():
                    setattr(o, key, value)
            obj = o
    return obj



from __future__ import annotations
import datetime
from enum import Enum
from typing import Dict


class Currency(Enum):
    EUR = "EUR"
    PLN = "PLN"
    NIS = "NIS"


class Order:
    def __init__(
        self,
        product: str = None,
        total: float = None,
        currency: Currency = None,
        quantity: int = None,
        region: int = None,
        at: datetime.datetime = None,
        tax: float = None,
    ):
        self.product = product
        self.total = total
        self.currency = currency
        self.quantity = quantity
        self.region = region
        self.at = at
        self.tax = tax

    @classmethod
    def from_json(cls, json_dict: Dict) -> Order:
        return cls(
            json_dict["product"],
            json_dict["total"],
            Currency(json_dict["currency"]),
            json_dict["quantity"],
            json_dict["region"],
            datetime.datetime.fromisoformat(json_dict["at"]) if "at" in json_dict else None,
            json_dict["tax"],
        )

    def to_json(self) -> Dict:
        return {
            "product": self.product,
            "total": self.total,
            "currency": self.currency.value,
            "quantity": self.quantity,
            "region": self.region,
            "at": self.at.isoformat() if self.at else None,
            "tax": self.tax,
        }



from __future__ import annotations

import json
from typing import Dict, Any, List, Optional

import requests

from ravendb import RavenCommand, RaftCommand, ServerNode
from ravendb.serverwide.operations.common import ServerOperation, T
from ravendb.util.util import RaftIdGenerator


class DisableDatabaseToggleResult:
    def __init__(self, disabled: bool = None, name: str = None, success: bool = None, reason: str = None) -> None:
        self.disabled = disabled
        self.name = name
        self.success = success
        self.reason = reason

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> "DisableDatabaseToggleResult":
        return cls(json_dict["Disabled"], json_dict["Name"], json_dict["Success"], json_dict["Reason"])


class ToggleDatabasesStateOperation(ServerOperation[DisableDatabaseToggleResult]):
    class Parameters:
        def __init__(self, databases_names: List[str] = None):
            self.databases_names = databases_names

        def to_json(self) -> Dict[str, Any]:
            return {"DatabaseNames": self.databases_names}

    def __init__(self, database_name: str, disable: bool):
        if database_name is None:
            raise ValueError("Database name cannot be None")

        self._disable = disable
        self._parameters = self.Parameters()
        self._parameters.databases_names = [database_name]

    @classmethod
    def from_multiple_names(cls, database_names: List[str], disable: bool):
        this = cls("", disable)
        this._parameters = cls.Parameters(database_names)
        return this

    @classmethod
    def from_parameters(cls, parameters: ToggleDatabasesStateOperation.Parameters, disable: bool):
        if parameters is None:
            raise ValueError("Parameters cannot be None")

        if not parameters.databases_names:
            raise ValueError("Parameters.DatabaseNames cannot be None or empty")

        this = cls("", disable)
        this._parameters = parameters
        return this

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[T]:
        return self.ToggleDatabaseStateCommand(self._parameters, self._disable)

    class ToggleDatabaseStateCommand(RavenCommand[DisableDatabaseToggleResult], RaftCommand):
        def __init__(self, parameters: ToggleDatabasesStateOperation.Parameters, disable: bool):
            super().__init__(DisableDatabaseToggleResult)

            if parameters is None:
                raise ValueError("Parameters cannot be None")

            self._parameters = parameters
            self._disable = disable

        def create_request(self, node: ServerNode) -> requests.Request:
            toggle = "disable" if self._disable else "enable"
            url = f"{node.url}/admin/databases/{toggle}"

            request = requests.Request(method="POST", url=url, data=self._parameters.to_json())

            return request

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            json_node = json.loads(response)
            status = json_node.get("Status")

            if not isinstance(status, list):
                self._throw_invalid_response()

            if not status:
                return None

            databases_status = status[0]
            self.result = DisableDatabaseToggleResult.from_json(databases_status)

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



import base64
import socket
import ssl
from typing import Tuple, Optional, Callable

from ravendb.documents.commands.subscriptions import TcpConnectionInfo
from ravendb.serverwide.tcp import TcpConnectionHeaderMessage


class TcpUtils:
    @staticmethod
    def connect(
        url_string: str,
        server_certificate_base64: Optional[str] = None,
        client_certificate_pem_path: Optional[str] = None,
        certificate_private_key_password: Optional[str] = None,
    ) -> socket.socket:
        hostname, port = url_string.replace("tcp://", "").split(":")
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        is_ssl_socket = server_certificate_base64 and client_certificate_pem_path
        if is_ssl_socket:
            context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
            context.load_cert_chain(client_certificate_pem_path, password=certificate_private_key_password)
            s = context.wrap_socket(s)

        s.connect((hostname, int(port)))
        if is_ssl_socket and base64.b64decode(server_certificate_base64) != s.getpeercert(True):
            raise ConnectionError("Failed to validate public server certificate.")
        return s

    @staticmethod
    def connect_with_priority(
        info: TcpConnectionInfo,
        server_cert_base64: Optional[str] = None,
        client_certificate_pem_path: Optional[str] = None,
        certificate_private_key_password: Optional[str] = None,
    ) -> Tuple[socket.socket, str]:
        if info.urls:
            for url in info.urls:
                try:
                    s = TcpUtils.connect(
                        url,
                        server_cert_base64,
                        client_certificate_pem_path,
                        certificate_private_key_password,
                    )
                    return s, url
                except Exception as e:
                    # ignored
                    pass

        s = TcpUtils.connect(
            info.url,
            server_cert_base64,
            client_certificate_pem_path,
            certificate_private_key_password,
        )

        return s, info.url

    @staticmethod
    def invoke_negotiation(
        info: TcpConnectionInfo,
        operation_type: TcpConnectionHeaderMessage.OperationTypes,
        negotiation_callback,
        url: str,
        socket: socket.socket,
    ) -> TcpConnectionHeaderMessage.SupportedFeatures:
        if operation_type == TcpConnectionHeaderMessage.OperationTypes.SUBSCRIPTION:
            return negotiation_callback(url, info, socket)
        else:
            raise NotImplementedError(f"Operation type '{operation_type}' not supported")

    class ConnectSecuredTcpSocketResult:
        def __init__(
            self,
            url: str = None,
            socket: socket.socket = None,
            supported_features: TcpConnectionHeaderMessage.SupportedFeatures = None,
        ):
            self.url = url
            self.socket = socket
            self.supported_features = supported_features

    @staticmethod
    def connect_secured_tcp_socket(
        info: TcpConnectionInfo,
        server_certificate: str,
        client_certificate_pem_path: str,
        certificate_private_key_password: Optional[str],
        operation_type: TcpConnectionHeaderMessage.OperationTypes,
        negotiation_callback: Callable,
    ) -> ConnectSecuredTcpSocketResult:
        if info.urls:
            for url in info.urls:
                try:
                    s = TcpUtils.connect(
                        url, server_certificate, client_certificate_pem_path, certificate_private_key_password
                    )
                    supported_features = TcpUtils.invoke_negotiation(info, operation_type, negotiation_callback, url, s)

                    return TcpUtils.ConnectSecuredTcpSocketResult(url, s, supported_features)
                except Exception as e:
                    pass
                    # ignored
        s = TcpUtils.connect(
            info.url, server_certificate, client_certificate_pem_path, certificate_private_key_password
        )

        supported_features = TcpUtils.invoke_negotiation(info, operation_type, negotiation_callback, info.url, s)
        return TcpUtils.ConnectSecuredTcpSocketResult(info.url, s, supported_features)



from typing import List, TYPE_CHECKING

import requests

from ravendb import SorterDefinition, RaftCommand, ServerNode
from ravendb.documents.operations.definitions import VoidMaintenanceOperation
from ravendb.http.raven_command import VoidRavenCommand
from ravendb.tools.utils import Utils
from ravendb.util.util import RaftIdGenerator

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class PutSortersOperation(VoidMaintenanceOperation):
    def __init__(self, *sorters_to_add: SorterDefinition):
        if not sorters_to_add:
            raise ValueError("Sorters must not be empty or None")

        self._sorters_to_add = sorters_to_add

    def get_command(self, conventions: "DocumentConventions") -> "PutSortersOperation.PutSortersCommand":
        return self.PutSortersCommand(conventions, list(self._sorters_to_add))

    class PutSortersCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, conventions: "DocumentConventions", sorters_to_add: List[SorterDefinition]):
            if conventions is None:
                raise ValueError("Conventions cannot be None")

            if sorters_to_add is None:
                raise ValueError("Sorters cannot be None")

            if any([sorter is None for sorter in sorters_to_add]):
                raise ValueError("Sorter cannot be None")

            super().__init__()

            self._sorters_to_add = sorters_to_add

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/admin/sorters"

            request = requests.Request("PUT", url)
            request.data = {"Sorters": [sorter.to_json() for sorter in self._sorters_to_add]}

            return request

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class DeleteSorterOperation(VoidMaintenanceOperation):
    def __init__(self, sorter_name: str = None):
        if sorter_name is None:
            raise ValueError("SorterName cannot be None")

        self._sorter_name = sorter_name

    def get_command(self, conventions: "DocumentConventions") -> "DeleteSorterOperation.DeleteSorterCommand":
        return self.DeleteSorterCommand(self._sorter_name)

    class DeleteSorterCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, sorter_name: str = None):
            if sorter_name is None:
                raise ValueError("SorterName cannot be None")

            super().__init__()
            self._sorter_name = sorter_name

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/admin/sorters?name={Utils.quote_key(self._sorter_name)}"
            return requests.Request("DELETE", url)

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



from __future__ import annotations
import logging
from typing import Optional, List, TYPE_CHECKING, Type, TypeVar, Set

from ravendb.documents.commands.crud import GetDocumentsCommand, GetDocumentsResult
from ravendb.documents.session.document_info import DocumentInfo
from ravendb.tools.utils import CaseInsensitiveSet, CaseInsensitiveDict, Utils

if TYPE_CHECKING:
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )
    from ravendb.documents.operations.time_series import TimeSeriesRange

_T = TypeVar("_T")


class LoadOperation:
    logger = logging.getLogger("load_operation")

    def __init__(
        self,
        session: InMemoryDocumentSessionOperations,
        keys: List[str] = None,
        includes: Optional[List[str]] = None,
        counters_to_include: Optional[List[str]] = None,
        compare_exchange_values_to_include: Optional[List[str]] = None,
        include_all_counters: Optional[bool] = None,
        time_series_to_include: Optional[List["TimeSeriesRange"]] = None,
    ):
        self._session = session
        self._keys = keys
        self._includes = includes or []
        self._counters_to_include = counters_to_include or []
        self._compare_exchange_values_to_include = compare_exchange_values_to_include or []

        self._time_series_to_include = time_series_to_include or []
        self._include_all_counters = include_all_counters
        self._time_series_to_include = time_series_to_include or []
        self._results_set = False
        self._results: Optional[GetDocumentsResult] = None

    def create_request(self) -> Optional[GetDocumentsCommand]:
        if self._session.check_if_id_already_included(
            self._keys, list(self._includes) if self._includes is not None else None
        ):
            return None

        self._session.increment_requests_count()

        self.logger.info(
            f"Requesting the following ids {','.join(self._keys)} from {self._session.advanced.store_identifier}"
        )

        if self._include_all_counters:
            return GetDocumentsCommand.from_multiple_ids_all_counters(
                self._keys,
                self._includes,
                True,
                self._time_series_to_include,
                self._compare_exchange_values_to_include,
                False,
            )

        return GetDocumentsCommand.from_multiple_ids(
            self._keys,
            self._includes,
            self._counters_to_include,
            self._time_series_to_include,
            self._compare_exchange_values_to_include,
            False,
        )

    def by_key(self, key: str):
        if not key:
            return self
        if self._keys is None:
            self._keys = [key]
        return self

    def with_includes(self, includes: Optional[Set[str]]):
        self._includes = includes
        return self

    def with_compare_exchange(self, cmpxch: List[str]):
        self._compare_exchange_values_to_include = cmpxch
        return self

    def with_counters(self, counters: List[str]):
        if counters:
            self._counters_to_include = counters
        return self

    def with_all_counters(self):
        self._include_all_counters = True
        return self

    def with_time_series(self, time_series: List["TimeSeriesRange"]):
        if time_series:
            self._time_series_to_include = time_series
        return self

    def by_keys(self, keys: List[Optional[str]]):
        distinct = CaseInsensitiveSet(filter(lambda x: x and not x.isspace(), keys))
        self._keys = list(distinct)
        return self

    def get_document(self, object_type: Type[_T]) -> _T:
        if self._session.no_tracking:
            if not self._results_set and len(self._keys) > 0:
                raise RuntimeError("Cannot execute get_document before operation execution.")

            if (not self._results) or not self._results.results:
                return None

            document = self._results.results[0]
            if not document:
                return None

            document_info = DocumentInfo.get_new_document_info(document)
            return self._session.track_entity_document_info(object_type, document_info)
        return self.__get_document(object_type, self._keys[0])

    def __get_document(self, object_type: Type[_T], key: str) -> _T:
        if key is None:
            # todo: fix these ugly protected calls below
            return Utils.get_default_value(object_type)

        if self._session.is_deleted(key):
            return Utils.get_default_value(object_type)

        doc = self._session._documents_by_id.get(key)
        if doc is not None:
            return self._session.track_entity_document_info(object_type, doc)

        doc = self._session._included_documents_by_id.get(key)
        if doc is not None:
            return self._session.track_entity_document_info(object_type, doc)

        return Utils.get_default_value(object_type)

    def get_documents(self, object_type: Type[_T]) -> CaseInsensitiveDict[str, _T]:
        final_results = CaseInsensitiveDict()
        if self._session.no_tracking:
            if (not self._results_set) and len(self._keys) > 0:
                raise ValueError("Cannot execute 'get_documents before operation execution.")

            for key in self._keys:
                if not key:
                    continue
                final_results[key] = None

            if (not self._results) or not self._results.results:
                return final_results

            for document in self._results.results:
                if not document:
                    continue
                new_document_info = DocumentInfo.get_new_document_info(document)
                final_results[new_document_info.key] = self._session.track_entity_document_info(
                    object_type, new_document_info
                )

            return final_results

        for key in self._keys:
            if not key:
                continue
            final_results[key] = self.__get_document(object_type, key)

        return final_results

    def set_result(self, result: GetDocumentsResult) -> None:
        self._results_set = True
        if self._session.no_tracking:
            self._results = result
            return

        if not result:
            self._session.register_missing(*self._keys)
            return

        self._session.register_includes(result.includes)

        if self._include_all_counters or self._counters_to_include:
            self._session.register_counters_ids_list(
                result.counter_includes, self._keys, self._counters_to_include, self._include_all_counters
            )

        if self._time_series_to_include:
            self._session.register_time_series(result.time_series_includes)

        if self._compare_exchange_values_to_include:
            self._session.cluster_transaction.register_compare_exchange_values(result.compare_exchange_includes)

        for document in result.results:
            if document is None:
                continue

            new_document_info = DocumentInfo.get_new_document_info(document)
            self._session._documents_by_id.add(new_document_info)

        for key in self._keys:
            value = self._session._documents_by_id.get(key, None)
            if value is None:
                self._session.register_missing(key)

        self._session.register_missing_includes(result.results, result.includes, self._includes)



from __future__ import annotations
from abc import abstractmethod
from typing import TypeVar, TYPE_CHECKING, Dict, List, Type, Optional, Set, Union

from ravendb.documents.store.lazy import Lazy

if TYPE_CHECKING:
    from ravendb.documents.session.document_session import DocumentSession

_T = TypeVar("_T")


class LoaderWithInclude:
    @abstractmethod
    def include(self, path: str) -> LoaderWithInclude:
        pass

    @abstractmethod
    def load(self, id_: str, object_type: Optional[Type[_T]] = None) -> _T:
        pass


class MultiLoaderWithInclude(LoaderWithInclude):
    def __init__(self, session: DocumentSession):
        self.__session = session
        self.__includes: Set[str] = set()

    def include(self, path: str) -> LoaderWithInclude:
        self.__includes.add(path)
        return self

    def load(self, id_or_ids: Union[List[str], str], object_type: Optional[Type[_T]] = None) -> _T:
        if not isinstance(id_or_ids, (str, list)):
            raise TypeError(f"Expected str or list of str, got '{type(id_or_ids)}'")
        ids = [id_or_ids] if isinstance(id_or_ids, str) else id_or_ids
        return self.__session._load_internal(object_type, ids, self.__includes)


class LazyMultiLoaderWithInclude(LoaderWithInclude):
    def __init__(self, session: "DocumentSession"):
        self.__session = session
        self.__includes = []

    def include(self, path: str) -> LazyMultiLoaderWithInclude:
        self.__includes.append(path)
        return self

    def load(self, id_or_ids: Union[List[str], str], object_type: Optional[Type[_T]] = None) -> Lazy[Dict[str, _T]]:
        if not isinstance(id_or_ids, (str, list)):
            raise TypeError(f"Expected str or list of str, got '{type(id_or_ids)}'")
        ids = [id_or_ids] if isinstance(id_or_ids, str) else id_or_ids
        return self.__session.lazy_load_internal(object_type, ids, self.__includes, None)



from __future__ import annotations

from typing import Dict, Any

from ravendb.documents.indexes.vector.embedding import VectorEmbeddingType
from ravendb.primitives import constants


class VectorOptions:
    def __init__(
        self,
        source_embedding_type: VectorEmbeddingType = constants.VectorSearch.DEFAULT_EMBEDDING_TYPE,
        destination_embedding_type: VectorEmbeddingType = constants.VectorSearch.DEFAULT_EMBEDDING_TYPE,
        dimensions: int = None,
        number_of_edges: int = None,
        number_of_candidates_for_indexing: int = None,
    ):
        self.dimensions = dimensions
        self.source_embedding_type = source_embedding_type
        self.destination_embedding_type = destination_embedding_type
        self.numbers_of_candidates_for_indexing = number_of_candidates_for_indexing
        self.number_of_edges = number_of_edges

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> VectorOptions:
        return cls(
            json_dict["SourceEmbeddingType"],
            json_dict["DestinationEmbeddingType"],
            json_dict["Dimensions"],
            json_dict["NumberOfEdges"],
            json_dict["NumberOfCandidatesForIndexing"],
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "SourceEmbeddingType": self.source_embedding_type.value,
            "DestinationEmbeddingType": self.destination_embedding_type.value,
            "Dimensions": self.dimensions,
            "NumberOfCandidatesForIndexing": self.numbers_of_candidates_for_indexing,
            "NumberOfEdges": self.number_of_edges,
        }


class AutoVectorOptions(VectorOptions):
    def __init__(
        self,
        source_embedding_type: VectorEmbeddingType = constants.VectorSearch.DEFAULT_EMBEDDING_TYPE,
        destination_embedding_type: VectorEmbeddingType = constants.VectorSearch.DEFAULT_EMBEDDING_TYPE,
        dimensions: int = None,
        number_of_edges: int = None,
        number_of_candidates_for_indexing: int = None,
        source_field_name: str = None,
    ):
        super().__init__(
            source_embedding_type,
            destination_embedding_type,
            dimensions,
            number_of_edges,
            number_of_candidates_for_indexing,
        )
        self.source_field_name = source_field_name

    @classmethod
    def from_vector_options(cls, vector_options: VectorOptions):
        return cls(
            source_embedding_type=vector_options.source_embedding_type,
            destination_embedding_type=vector_options.destination_embedding_type,
            dimensions=vector_options.dimensions,
            number_of_edges=vector_options.number_of_edges,
            number_of_candidates_for_indexing=vector_options.numbers_of_candidates_for_indexing,
        )

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]):
        vec_options = super().from_json(json_dict)
        auto_vect_options = cls.from_vector_options(vec_options)
        auto_vect_options.source_field_name = json_dict["SourceFieldName"]
        return auto_vect_options

    def to_json(self) -> Dict[str, Any]:
        json_dict = super().to_json()
        json_dict["SourceFieldName"] = self.source_field_name
        return json_dict
